import re
from typing import Tuple

from .. import Params, Parseable
from ..exceptions import NotParseable

__all__ = ['Tag']


class Tag(Parseable[bytes]):
    """Represents the tag prefixed to every client command in an IMAP stream.

    Args:
        tag: The tag value.

    """

    _pattern = re.compile(br'[\x21\x23\x24\x26\x27\x2C-\x5B'
                          br'\x5D\x5E-\x7A\x7C\x7E]+')

    def __init__(self, tag: bytes) -> None:
        super().__init__()
        self.tag = tag

    @property
    def value(self) -> bytes:
        """The tag value."""
        return self.tag

    @classmethod
    def parse(cls, buf: bytes, params: Params) -> Tuple['Tag', bytes]:
        start = cls._whitespace_length(buf)
        match = cls._pattern.match(buf, start)
        if not match:
            raise NotParseable(buf)
        return cls(match.group(0)), buf[match.end(0):]

    def __bytes__(self) -> bytes:
        return self.value



"""Defines a configuration that uses an in-memory dictionary for example usage
and integration testing.

"""

import os.path
from argparse import Namespace, ArgumentDefaultsHelpFormatter
from contextlib import closing
from datetime import datetime, timezone
from typing import cast, Any, Tuple, Mapping, Dict, BinaryIO, TypeVar, Type

from pkg_resources import resource_listdir, resource_stream
from pysasl import AuthenticationCredentials

from pymap.config import IMAPConfig
from pymap.exceptions import InvalidAuth
from pymap.interfaces.session import LoginProtocol
from pymap.parsing.specials.flag import Flag, Recent

from .mailbox import Message, MailboxData, MailboxSet
from ..session import BaseSession

__all__ = ['add_subparser', 'init', 'Config', 'Session']

_SessionT = TypeVar('_SessionT', bound='Session')


def add_subparser(subparsers) -> None:
    parser = subparsers.add_parser(
        'dict', help='in-memory backend',
        formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('-d', '--demo-data', action='store_true',
                        help='load initial demo data')
    parser.add_argument('-u', '--demo-user', default='demouser',
                        metavar='VAL', help='demo user ID')
    parser.add_argument('-p', '--demo-password', default='demopass',
                        metavar='VAL', help='demo user password')


async def init(args: Namespace) -> Tuple[LoginProtocol, 'Config']:
    return Session.login, Config.from_args(args)


class Config(IMAPConfig):
    """The config implementation for the dict backend.

    Args:
        args: The command-line arguments.
        demo_data: Load the demo data, used by integration tests.

    """

    def __init__(self, args: Namespace, demo_data: bool, **extra: Any) -> None:
        super().__init__(args, **extra)
        self.demo_data = demo_data
        self.set_cache: Dict[str, MailboxSet] = {}

    @property
    def demo_user(self) -> str:
        """Used by the default :meth:`~Session.get_password` implementation
        to retrieve the ``--demo-user`` command-line argument, which defaults
        to ``demouser``.

        """
        return self.args.demo_user

    @property
    def demo_password(self) -> str:
        """Used by the default :meth:`~Session.get_password` implementation
        to retrieve the ``--demo-password`` command-line argument, which
        defaults to ``demopass``.

        """
        return self.args.demo_password

    @classmethod
    def parse_args(cls, args: Namespace, **extra: Any) -> Mapping[str, Any]:
        return super().parse_args(args, demo_data=args.demo_data, **extra)


class Session(BaseSession[MailboxData]):
    """The session implementation for the dict backend."""

    resource = __name__

    @classmethod
    async def login(cls: Type[_SessionT],
                    credentials: AuthenticationCredentials,
                    config: Config) -> _SessionT:
        """Checks the given credentials for a valid login and returns a new
        session. The mailbox data is shared between concurrent and future
        sessions, but only for the lifetime of the process.

        """
        user = credentials.authcid
        password = await cls.get_password(config, user)
        if not password or not credentials.check_secret(password):
            raise InvalidAuth()
        mailbox_set = config.set_cache.get(user)
        if not mailbox_set:
            mailbox_set = MailboxSet()
            if config.demo_data:
                await cls._load_demo(mailbox_set)
            config.set_cache[user] = mailbox_set
        return cls(mailbox_set)

    @classmethod
    async def get_password(cls, config: Config, user: str) -> str:
        """If the given user ID exists, return its expected password. Override
        this method to implement custom login logic.

        Args:
            config: The dict config object.
            user: The expected user ID.

        Raises:
            InvalidAuth: The user ID was not valid.

        """
        expected_user: str = config.demo_user
        expected_password: str = config.demo_password
        if user == expected_user:
            return expected_password
        raise InvalidAuth()

    @classmethod
    async def _load_demo(cls, mailbox_set: MailboxSet) -> None:
        inbox = await mailbox_set.get_mailbox('INBOX')
        await cls._load_demo_mailbox('INBOX', inbox)
        mbx_names = sorted(resource_listdir(cls.resource, 'demo'))
        for name in mbx_names:
            if name != 'INBOX':
                mbx = await mailbox_set.add_mailbox(name)
                await cls._load_demo_mailbox(name, mbx)

    @classmethod
    async def _load_demo_mailbox(cls, name: str, mbx: MailboxData) -> None:
        path = os.path.join('demo', name)
        msg_names = sorted(resource_listdir(cls.resource, path))
        for msg_name in msg_names:
            if msg_name == '.readonly':
                mbx._readonly = True
                continue
            elif msg_name.startswith('.'):
                continue
            msg_path = os.path.join(path, msg_name)
            message_stream = resource_stream(cls.resource, msg_path)
            with closing(message_stream):
                msg_file = cast(BinaryIO, message_stream)
                flags_line = message_stream.readline()
                msg_timestamp = float(message_stream.readline())
                msg_dt = datetime.fromtimestamp(msg_timestamp, timezone.utc)
                msg_flags = {Flag(flag) for flag in flags_line.split()}
                if Recent in msg_flags:
                    msg_flags.remove(Recent)
                    msg_recent = True
                else:
                    msg_recent = False
                msg = Message.parse(0, msg_file, msg_flags, msg_dt)
            await mbx.add(msg, msg_recent)




import unittest

from pymap.parsing.specials import SequenceSet


class TestSequenceSet(unittest.TestCase):

    def test_contains(self):
        set1 = SequenceSet([12])
        self.assertTrue(set1.contains(12, 100))
        self.assertFalse(set1.contains(13, 100))
        self.assertFalse(set1.contains(12, 10))
        set2 = SequenceSet(['*'])
        self.assertTrue(set2.contains(100, 100))
        self.assertFalse(set2.contains(21, 100))
        self.assertFalse(set2.contains(21, 20))
        set3 = SequenceSet([(5, 10)])
        self.assertTrue(set3.contains(7, 100))
        self.assertFalse(set3.contains(7, 6))
        self.assertFalse(set3.contains(11, 100))
        self.assertFalse(set3.contains(4, 100))
        set4 = SequenceSet([('*', 10)])
        self.assertTrue(set4.contains(11, 100))
        self.assertFalse(set4.contains(9, 100))
        self.assertFalse(set4.contains(101, 100))
        set5 = SequenceSet([(10, '*')])
        self.assertTrue(set5.contains(11, 100))
        self.assertFalse(set5.contains(9, 100))
        self.assertFalse(set5.contains(101, 100))
        set6 = SequenceSet([('*', '*')])
        self.assertTrue(set6.contains(12, 12))
        self.assertFalse(set6.contains(13, 12))
        self.assertFalse(set6.contains(11, 12))

    def test_iter(self):
        set1 = SequenceSet([12])
        self.assertEqual([12], list(set1.iter(100)))
        set2 = SequenceSet(['*'])
        self.assertEqual([100], list(set2.iter(100)))
        set3 = SequenceSet([('*', '*')])
        self.assertEqual([100], list(set3.iter(100)))
        set4 = SequenceSet([(5, 10)])
        self.assertEqual(list(range(5, 11)), list(set4.iter(100)))
        set5 = SequenceSet([('*', 10)])
        self.assertEqual(list(range(10, 101)), list(set5.iter(100)))
        set6 = SequenceSet([(10, '*')])
        self.assertEqual(list(range(10, 101)), list(set6.iter(100)))
        set7 = SequenceSet([(1, 2), (4, 5), (7, 8)])
        self.assertEqual([1, 2, 4, 5, 7, 8], list(set7.iter(100)))
        set8 = SequenceSet([(3, 5), (4, 6)])
        self.assertEqual([3, 4, 5, 6], list(set8.iter(100)))
        set9 = SequenceSet([('*', 50), (60, 10)])
        self.assertEqual(list(range(10, 101)), list(set9.iter(100)))
        set10 = SequenceSet([(10, 1000)])
        self.assertEqual(list(range(10, 101)), list(set10.iter(100)))
        set11 = SequenceSet([(1000, 1000)])
        self.assertEqual([], list(set11.iter(100)))

    def test_bytes(self):
        seq = SequenceSet([12, '*', (1, '*')])
        self.assertEqual(b'12,*,1:*', bytes(seq))
        self.assertEqual(b'12,*,1:*', bytes(seq))

    def test_build(self):
        seq = SequenceSet.build([1])
        self.assertEqual(b'1', bytes(seq))
        seq = SequenceSet.build([1, 1, 1])
        self.assertEqual(b'1', bytes(seq))
        seq = SequenceSet.build([1, 2, 3, 4, 5])
        self.assertEqual(b'1:5', bytes(seq))
        seq = SequenceSet.build([1, 2, 4, 5])
        self.assertEqual(b'1:2,4:5', bytes(seq))
        seq = SequenceSet.build([1, 2, 4, 6])
        self.assertEqual(b'1:2,4,6', bytes(seq))
        seq = SequenceSet.build([1, 3, 5])
        self.assertEqual(b'1,3,5', bytes(seq))
        seq = SequenceSet.build([1, 2, 3, 4, 5])



"""Contains the context variables that are made available by the IMAP server to
backend implementations.

Attributes:
    current_command: The currently executing command.
    socket_info: Info about the currently connected socket.

"""

from contextvars import ContextVar

from .parsing.command import Command
from .sockinfo import SocketInfo

__all__ = ['current_command', 'socket_info']

current_command: ContextVar[Command] = ContextVar('current_command')
socket_info: ContextVar[SocketInfo] = ContextVar('socket_info')



"""Package defining all the IMAP parsing and response classes."""

import re
from abc import abstractmethod, ABCMeta
from typing import Tuple, Generic, TypeVar, Sequence, Dict, Any, List, Type

from .exceptions import NotParseable, UnexpectedType

__all__ = ['Params', 'Parseable', 'ExpectedParseable', 'Space', 'EndLine',
           'ParseableT']

#: Type variable used for specifying the parseable type of a :class:`Parseable`
#: sub-class.
ParseableT = TypeVar('ParseableT')


class Params:
    """Parameters used and passed around among the :meth:`~Parseable.parse`
    methods.

    Args:
        continuations: The continuation buffers remaining for parsing.
        expected: The types that are expected in the next parsed object.
        list_expected: The types that are expect in a parsed list.
        command_name: The name of the command currently being parsed, if any.
        uid: The next parsed command came after a ``UID`` command.
        charset: Strings should be decoded using this character set.
        tag: The next parsed command uses this tag bytestring.
        max_append_len: The maximum allowed length of the message body to an
            ``APPEND`` command.

    """

    __slots__ = ['continuations', 'expected', 'list_expected', 'command_name',
                 'uid', 'charset', 'tag', 'max_append_len']

    def __init__(self, continuations: List[bytes] = None,
                 expected: Sequence[Type['Parseable']] = None,
                 list_expected: Sequence[Type['Parseable']] = None,
                 command_name: bytes = None,
                 uid: bool = False,
                 charset: str = None,
                 tag: bytes = None,
                 max_append_len: int = None) -> None:
        super().__init__()
        self.continuations = continuations or []
        self.expected = expected or []
        self.list_expected = list_expected or []
        self.command_name = command_name
        self.uid = uid
        self.charset = charset
        self.tag = tag or b'.'
        self.max_append_len = max_append_len

    def _set_if_none(self, kwargs: Dict[str, Any], attr: str, value) -> None:
        if value is not None:
            kwargs[attr] = value
        else:
            kwargs[attr] = getattr(self, attr)

    def copy(self, continuations: List[bytes] = None,
             expected: Sequence[Type['Parseable']] = None,
             list_expected: Sequence[Type['Parseable']] = None,
             command_name: bytes = None,
             uid: bool = None,
             charset: str = None,
             tag: bytes = None,
             max_append_len: int = None) -> 'Params':
        """Copy the parameters, possibly replacing a subset."""
        kwargs: Dict[str, Any] = {}
        self._set_if_none(kwargs, 'continuations', continuations)
        self._set_if_none(kwargs, 'expected', expected)
        self._set_if_none(kwargs, 'list_expected', list_expected)
        self._set_if_none(kwargs, 'command_name', command_name)
        self._set_if_none(kwargs, 'uid', uid)
        self._set_if_none(kwargs, 'charset', charset)
        self._set_if_none(kwargs, 'tag', tag)
        self._set_if_none(kwargs, 'max_append_len', max_append_len)
        return Params(**kwargs)


class Parseable(Generic[ParseableT], metaclass=ABCMeta):
    """Represents a parseable data object from an IMAP stream. The sub-classes
    implement the different data formats.

    This base class will be inherited by all necessary entries in the IMAP
    formal syntax section.

    """

    _whitespace_pattern = re.compile(br' +')
    _atom_pattern = re.compile(br'[\x21\x23\x24\x26\x27\x2B'
                               br'-\x5B\x5E-\x7A\x7C\x7E]+')

    __slots__ = []  # type: ignore

    @property
    @abstractmethod
    def value(self) -> ParseableT:
        """The primary value associated with the parsed data."""
        ...

    @classmethod
    def _whitespace_length(cls, buf: bytes, start: int = 0) -> int:
        match = cls._whitespace_pattern.match(buf, start)
        if match:
            return match.end(0) - start
        return 0

    @abstractmethod
    def __bytes__(self) -> bytes:
        ...

    def __eq__(self, other) -> bool:
        if isinstance(other, bytes):
            return bytes(self) == other
        return NotImplemented

    @classmethod
    @abstractmethod
    def parse(cls, buf: bytes, params: 'Params') -> Tuple['Parseable', bytes]:
        """Implemented by sub-classes to define how to parse the given buffer.

        Args:
            buf: The bytes containing the data to be parsed.
            params: The parameters used by some parseable types.

        """
        ...


class ExpectedParseable(Parseable[None]):
    """Non-instantiable class, used to parse a buffer from a list of
    expected types.

    """

    def __init__(self) -> None:
        super().__init__()
        raise NotImplementedError

    @property
    def value(self) -> None:
        raise NotImplementedError

    def __bytes__(self) -> bytes:
        raise NotImplementedError

    @classmethod
    def parse(cls, buf: bytes, params: 'Params') -> Tuple['Parseable', bytes]:
        """Parses the given buffer by attempting to parse the list of
        :attr:`~Params.expected` types until one of them succeeds,
        then returns the parsed object.

        Args:
            buf: The bytes containing the data to be parsed.
            params: The parameters used by some parseable types.

        """
        for data_type in params.expected:
            try:
                return data_type.parse(buf, params)
            except NotParseable:
                pass
        raise UnexpectedType(buf)


class Space(Parseable[int]):
    """Represents at least one space character.

    Args:
        length: The number of spaces parsed.

    """

    __slots__ = ['length']

    def __init__(self, length: int) -> None:
        super().__init__()
        self.length = length

    @property
    def value(self) -> int:
        """The number of spaces parsed."""
        return self.length

    @classmethod
    def parse(cls, buf: bytes, params: 'Params') -> Tuple['Space', bytes]:
        ret = cls._whitespace_length(buf)
        if not ret:
            raise NotParseable(buf)
        return cls(ret), buf[ret:]

    def __bytes__(self) -> bytes:
        return b' ' * self.length


class EndLine(Parseable[bytes]):
    """Represents the end of a parsed line. This will only parse if the buffer
    has zero or more space characters followed by a new-line sequence.

    Args:
        preceding_spaces: The number of space characteres before the newline.
        carriage_return: Whether the newline included a carriage return.

    Attributes:
        preceding_spaces: The number of space characteres before the newline.
        carriage_return: Whether the newline included a carriage return.

    """

    _pattern = re.compile(br' *(\r?)\n')

    __slots__ = ['preceding_spaces', 'carriage_return']

    def __init__(self, preceding_spaces: int = 0,
                 carriage_return: bool = True) -> None:
        super().__init__()
        self.preceding_spaces = preceding_spaces
        self.carriage_return = carriage_return

    @property
    def value(self) -> bytes:
        """The endline bytestring."""
        return b'\r\n' if self.carriage_return else b'\n'

    @classmethod
    def parse(cls, buf: bytes, params: 'Params') -> Tuple['EndLine', bytes]:
        match = cls._pattern.match(buf)
        if not match:
            raise NotParseable(buf)
        preceding_spaces = match.start(1)
        carriage_return = bool(match.group(1))
        return cls(preceding_spaces, carriage_return), buf[match.end(0):]

    def __bytes__(self) -> bytes:
        return b' ' * self.preceding_spaces + self.value




from abc import abstractmethod
from typing import TypeVar, Optional, Tuple, Sequence, FrozenSet, \
    Iterable, AsyncIterable
from typing_extensions import Protocol

from pymap.mailbox import MailboxSnapshot
from pymap.message import AppendMessage, BaseMessage
from pymap.parsing.specials import SequenceSet, FetchRequirement
from pymap.parsing.specials.flag import get_system_flags, Flag, Recent, Seen
from pymap.selected import SelectedSet, SelectedMailbox

from .util import asyncenumerate

__all__ = ['MailboxDataInterface', 'MailboxSetInterface', 'Message',
           'MessageT', 'MailboxDataT_co']

#: Type variable with an upper bound of :class:`Message`.
MessageT = TypeVar('MessageT', bound='Message')

#: Covariant type variable with an upper bound of
#: :class:`MailboxDataInterface`.
MailboxDataT_co = TypeVar('MailboxDataT_co', bound='MailboxDataInterface',
                          covariant=True)


class Message(BaseMessage):
    """Manages a single message. This message does not have its contents
    (headers, body, etc.) loaded into memory, it is only the IMAP metadata
    such as UID and flags.

    """

    @property
    def recent(self) -> bool:
        """True if the message is considered new in the mailbox. The next
        session to SELECT the mailbox will negate this value and apply the
        ``\\Recent`` session flag to the message.

        """
        return self._kwargs.get('recent', False)

    @recent.setter
    def recent(self, recent: bool) -> None:
        self._kwargs['recent'] = recent


class MailboxDataInterface(Protocol[MessageT]):
    """Manages the messages and metadata associated with a single mailbox."""

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the mailbox."""
        ...

    @property
    @abstractmethod
    def readonly(self) -> bool:
        """Whether the mailbox is read-only or read-write."""
        ...

    @property
    @abstractmethod
    def uid_validity(self) -> int:
        """The mailbox UID validity value."""
        ...

    @property
    @abstractmethod
    def next_uid(self) -> int:
        """The predicted next message UID."""
        ...

    @property
    def permanent_flags(self) -> FrozenSet[Flag]:
        """The permanent flags allowed in the mailbox."""
        return get_system_flags() - {Recent}

    @property
    def session_flags(self) -> FrozenSet[Flag]:
        """The session flags allowed in the mailbox."""
        return frozenset({Recent})

    @property
    @abstractmethod
    def selected_set(self) -> SelectedSet:
        """The set of selected mailbox sessions currently active."""
        ...

    @abstractmethod
    def parse_message(self, append_msg: AppendMessage) -> MessageT:
        """Parse the raw message data into a loaded message object.

        Args:
            append_msg: A single message from the APPEND command.

        """
        ...

    @abstractmethod
    async def add(self, message: MessageT, recent: bool = False) -> MessageT:
        """Adds a new message to the end of the mailbox, returning a copy of
        message with its assigned UID.

        Args:
            message: The loaded message object.
            recent: True if the message should be marked recent.

        """
        ...

    @abstractmethod
    async def get(self, uid: int,
                  requirement: FetchRequirement = FetchRequirement.METADATA) \
            -> Optional[MessageT]:
        """Return the message with the given UID. If the UID has been
        expunged, a message with
        :attr:`~pymap.interfaces.message.MessageInterface.expunged` set to
        True should be returned.

        Args:
            uid: The message UID.
            requirement: The data required from each message.

        """
        ...

    @abstractmethod
    async def delete(self, uids: Iterable[int]) -> None:
        """Delete messages with the given UIDs.

        Args:
            uids: The message UIDs.

        """
        ...

    @abstractmethod
    async def claim_recent(self, selected: SelectedMailbox) -> None:
        """Messages that are newly added to the mailbox are assigned the
        ``\\Recent`` flag in the current selected mailbox session.

        Args:
            selected: The selected mailbox session.

        """
        ...

    @abstractmethod
    async def save_flags(self, messages: Iterable[MessageT]) -> None:
        """Save the flags currently stored in each message's
        :attr:`~pymap.interfaces.message.Message.permanent_flags` set.

        Args:
            messages: The message objects.

        """
        ...

    @abstractmethod
    async def cleanup(self) -> None:
        """Perform any necessary "housekeeping" steps. This may be a slow
        operation, and may run things like garbage collection on the backend.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.check_mailbox`

        """
        ...

    @abstractmethod
    def uids(self) -> AsyncIterable[int]:
        """Return all of the active message UIDs in the mailbox."""
        ...

    @abstractmethod
    def messages(self) -> AsyncIterable[MessageT]:
        """Return all of the active messages in the mailbox."""
        ...

    @abstractmethod
    def items(self) -> AsyncIterable[Tuple[int, MessageT]]:
        """Return all of the active message UID and message pairs in the
        mailbox.

        """
        ...

    async def find(self, seq_set: SequenceSet, selected: SelectedMailbox,
                   requirement: FetchRequirement = FetchRequirement.METADATA) \
            -> AsyncIterable[Tuple[int, MessageT]]:
        """Find the active message UID and message pairs in the mailbox that
        are contained in the given sequences set. Message sequence numbers
        are resolved by the selected mailbox session.

        Args:
            seq_set: The sequence set of the desired messages.
            selected: The selected mailbox session.
            requirement: The data required from each message.

        """
        mbx_uids = frozenset({uid async for uid in self.uids()})
        for seq, uid in selected.iter_set(seq_set, mbx_uids):
            msg = await self.get(uid, requirement)
            if msg is not None:
                yield (seq, msg)

    async def snapshot(self) -> MailboxSnapshot:
        """Returns a snapshot of the current state of the mailbox."""
        exists = 0
        recent = 0
        unseen = 0
        first_unseen: Optional[int] = None
        async for seq, msg in asyncenumerate(self.messages(), 1):
            exists += 1
            if msg.recent:
                recent += 1
            if Seen not in msg.permanent_flags:
                unseen += 1
                if first_unseen is None:
                    first_unseen = seq
        return MailboxSnapshot(self.name, self.readonly, self.uid_validity,
                               self.permanent_flags, self.session_flags,
                               exists, recent, unseen, first_unseen,
                               self.next_uid)


class MailboxSetInterface(Protocol[MailboxDataT_co]):
    """Manages the set of mailboxes available to the authenticated user."""

    @property
    @abstractmethod
    def inbox(self) -> MailboxDataT_co:
        """The INBOX mailbox, which must always exist."""
        ...

    @property
    @abstractmethod
    def delimiter(self) -> str:
        """The delimiter used in mailbox names to indicate hierarchy."""
        ...

    @abstractmethod
    async def set_subscribed(self, name: str, subscribed: bool) -> None:
        """Add or remove the subscribed status of a mailbox.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.subscribe`
            :meth:`~pymap.interfaces.session.SessionInterface.unsubscribe`

        Args:
            name: The name of the mailbox.
            subscribed: True if the mailbox should be subscribed.

        """
        ...

    @abstractmethod
    async def list_subscribed(self) -> Sequence[str]:
        """Return a list of all subscribed mailboxes.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.list_mailboxes`

        """
        ...

    @abstractmethod
    async def list_mailboxes(self) -> Sequence[str]:
        """Return a list of all mailboxes.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.list_mailboxes`

        """
        ...

    @abstractmethod
    async def get_mailbox(self, name: str, try_create: bool = False) \
            -> MailboxDataT_co:
        """Return an existing mailbox.

        Args:
            name: The name of the mailbox.
            try_create: True if the operation might succeed if the mailbox
                is created first.

        Raises:
            :exc:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def add_mailbox(self, name: str) -> MailboxDataT_co:
        """Create a new mailbox.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.create_mailbox`

        Args:
            name: The name of the mailbox.

        Raises:
            :exc:`~pymap.exceptions.MailboxConflict`

        """
        ...

    @abstractmethod
    async def delete_mailbox(self, name: str) -> None:
        """Delete an existing mailbox.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.delete_mailbox`

        Args:
            name: The name of the mailbox.

        Raises:
            :exc:`~pymap.exceptions.MailboxNotFound`
            :exc:`~pymap.exceptions.MailboxHasChildren`

        """
        ...

    @abstractmethod
    async def rename_mailbox(self, before: str, after: str) -> MailboxDataT_co:
        """Rename an existing mailbox.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.rename_mailbox`

        Args:
            before: The name of the existing mailbox.
            after: The name of the destination mailbox.

        Raises:

        """
        ...



import re
from datetime import datetime
from typing import TypeVar, Type, Tuple, Union, Sequence, FrozenSet

from .astring import AString
from .fetchattr import FetchRequirement
from .flag import Flag
from .sequenceset import SequenceSet
from .. import Params, Parseable, ExpectedParseable, Space
from ..exceptions import NotParseable, UnexpectedType
from ..primitives import Atom, Number, QuotedString, ListP

__all__ = ['SearchKey']

_FilterT = TypeVar('_FilterT', bound='_FilterType')
_FilterType = Union[Tuple['SearchKey', 'SearchKey'], Tuple[str, str],
                    Sequence['SearchKey'], SequenceSet, Flag,
                    datetime, int, str]


class SearchKey(Parseable[bytes]):
    """Represents a search key given to the SEARCH command on an IMAP stream.

    Args:
        key: The search key.
        filter_: The filter object, used by most search keys.
        inverse: If the search key was inverted with ``NOT``.

    Args:
        filter_: The filter object, used by most search keys.
        inverse: If the search key was inverted with ``NOT``.

    """

    _not_pattern = re.compile(br'NOT +', re.I)

    def __init__(self, key: bytes,
                 filter_: _FilterType = None,
                 inverse: bool = False) -> None:
        super().__init__()
        self.key = key
        self.filter = filter_
        self.inverse = inverse

    @property
    def value(self) -> bytes:
        """The search key."""
        return self.key

    @property
    def requirement(self) -> FetchRequirement:
        """Indicates the data required to fulfill this search key."""
        key_name = self.key
        if key_name == b'ALL':
            return FetchRequirement.NONE
        elif key_name == b'KEYSET':
            keyset_reqs = {key.requirement for key in self.filter_key_set}
            return FetchRequirement.reduce(keyset_reqs)
        elif key_name == b'OR':
            left, right = self.filter_key_or
            key_or_reqs = {left.requirement, right.requirement}
            return FetchRequirement.reduce(key_or_reqs)
        elif key_name in (b'SENTBEFORE', b'SENTON', b'SENTSINCE', b'BCC',
                          b'CC', b'FROM', b'SUBJECT', b'TO', b'HEADER'):
            return FetchRequirement.HEADERS
        elif key_name in (b'BODY', b'TEXT'):
            return FetchRequirement.BODY
        else:
            return FetchRequirement.METADATA

    @property
    def not_inverse(self) -> 'SearchKey':
        """Return a copy of the search key with :attr:`.inverse` flipped."""
        return SearchKey(self.value, self.filter, not self.inverse)

    def _get_filter(self, cls: Type[_FilterT]) -> _FilterT:
        if not isinstance(self.filter, cls):
            raise TypeError(self.filter)
        return self.filter

    @property
    def filter_sequence_set(self) -> SequenceSet:
        return self._get_filter(SequenceSet)

    @property
    def filter_key_set(self) -> FrozenSet['SearchKey']:
        return self._get_filter(frozenset)  # type: ignore

    @property
    def filter_key_or(self) -> Tuple['SearchKey', 'SearchKey']:
        return self._get_filter(tuple)  # type: ignore

    @property
    def filter_flag(self) -> Flag:
        return self._get_filter(Flag)

    @property
    def filter_datetime(self) -> datetime:
        return self._get_filter(datetime)

    @property
    def filter_int(self) -> int:
        return self._get_filter(int)

    @property
    def filter_str(self) -> str:
        return self._get_filter(str)

    @property
    def filter_header(self) -> Tuple[str, str]:
        return self._get_filter(tuple)  # type: ignore

    def __bytes__(self) -> bytes:
        raise NotImplementedError

    def __hash__(self) -> int:
        return hash((self.value, self.filter, self.inverse))

    def __eq__(self, other) -> bool:
        if isinstance(other, SearchKey):
            return hash(self) == hash(other)
        return super().__eq__(other)

    def __ne__(self, other) -> bool:
        if isinstance(other, SearchKey):
            return hash(self) != hash(other)
        return super().__ne__(other)

    @classmethod
    def _parse_astring_filter(cls, buf: bytes, params: Params):
        ret, after = AString.parse(buf, params)
        return ret.value.decode(params.charset or 'ascii'), after

    @classmethod
    def _parse_date_filter(cls, buf: bytes, params: Params):
        params_copy = params.copy(expected=[Atom, QuotedString])
        atom, after = ExpectedParseable.parse(buf, params_copy)
        date_str = str(atom.value, 'ascii', 'ignore')
        try:
            date = datetime.strptime(date_str, '%d-%b-%Y')
        except ValueError:
            raise NotParseable(buf)
        return date, after

    @classmethod
    def parse(cls, buf: bytes, params: Params) -> Tuple['SearchKey', bytes]:
        try:
            _, buf = Space.parse(buf, params)
        except NotParseable:
            pass
        inverse = False
        match = cls._not_pattern.match(buf)
        if match:
            inverse = True
            buf = buf[match.end(0):]
        try:
            seq_set, buf = SequenceSet.parse(buf, params)
        except NotParseable:
            pass
        else:
            return cls(b'SEQSET', seq_set, inverse), buf
        try:
            params_copy = params.copy(list_expected=[SearchKey])
            key_list_p, buf = ListP.parse(buf, params_copy)
        except UnexpectedType:
            raise
        except NotParseable:
            pass
        else:
            key_list = key_list_p.get_as(SearchKey)
            return cls(b'KEYSET', key_list, inverse), buf
        atom, after = Atom.parse(buf, params)
        key = atom.value.upper()
        if key in (b'ALL', b'ANSWERED', b'DELETED', b'FLAGGED', b'NEW', b'OLD',
                   b'RECENT', b'SEEN', b'UNANSWERED', b'UNDELETED',
                   b'UNFLAGGED', b'UNSEEN', b'DRAFT', b'UNDRAFT'):
            return cls(key, inverse=inverse), after
        elif key in (
                b'BCC', b'BODY', b'CC', b'FROM', b'SUBJECT', b'TEXT', b'TO'):
            _, buf = Space.parse(after, params)
            filter_, buf = cls._parse_astring_filter(buf, params)
            return cls(key, filter_, inverse), buf
        elif key in (b'BEFORE', b'ON', b'SINCE', b'SENTBEFORE', b'SENTON',
                     b'SENTSINCE'):
            _, buf = Space.parse(after, params)
            filter_, buf = cls._parse_date_filter(buf, params)
            return cls(key, filter_, inverse), buf
        elif key in (b'KEYWORD', b'UNKEYWORD'):
            _, buf = Space.parse(after, params)
            flag, after_buf = Flag.parse(buf, params)
            if flag.is_system:
                raise NotParseable(buf)
            return cls(key, flag, inverse), after_buf
        elif key in (b'LARGER', b'SMALLER'):
            _, buf = Space.parse(after, params)
            num, buf = Number.parse(buf, params)
            return cls(key, num.value, inverse), buf
        elif key == b'UID':
            _, buf = Space.parse(after, params)
            seq_set, buf = SequenceSet.parse(buf, params.copy(uid=True))
            return cls(b'SEQSET', seq_set, inverse), buf
        elif key == b'HEADER':
            _, buf = Space.parse(after, params)
            header_field, buf = cls._parse_astring_filter(buf, params)
            _, buf = Space.parse(buf, params)
            header_value, buf = cls._parse_astring_filter(buf, params)
            return cls(key, (header_field, header_value), inverse), buf
        elif key == b'OR':
            _, buf = Space.parse(after, params)
            or1, buf = SearchKey.parse(buf, params)
            _, buf = Space.parse(buf, params)
            or2, buf = SearchKey.parse(buf, params)
            return cls(key, (or1, or2), inverse), buf
        raise NotParseable(buf)




import unittest

from pymap.parsing.response.code import Alert, Capability, Parse, \
    PermanentFlags, ReadOnly, ReadWrite, TryCreate, UidNext, UidValidity, \
    Unseen, AppendUid, CopyUid


class TestAlert(unittest.TestCase):

    def test_bytes(self):
        code = Alert()
        self.assertEqual(b'[ALERT]', bytes(code))


class TestCapability(unittest.TestCase):

    def test_bytes(self):
        code = Capability([])
        self.assertEqual(b'[CAPABILITY IMAP4rev1]', bytes(code))
        code = Capability([b'TEST'])
        self.assertEqual(b'[CAPABILITY IMAP4rev1 TEST]', bytes(code))
        code = Capability([b'TEST', b'STUFF'])
        self.assertEqual(b'[CAPABILITY IMAP4rev1 TEST STUFF]', bytes(code))


class TestParse(unittest.TestCase):

    def test_bytes(self):
        code = Parse()
        self.assertEqual(b'[PARSE]', bytes(code))


class TestPermanentFlags(unittest.TestCase):

    def test_bytes(self):
        code = PermanentFlags([br'\One', br'\Two'])
        self.assertEqual(br'[PERMANENTFLAGS (\One \Two)]', bytes(code))


class TestReadOnly(unittest.TestCase):

    def test_bytes(self):
        code = ReadOnly()
        self.assertEqual(b'[READ-ONLY]', bytes(code))


class TestReadWrite(unittest.TestCase):

    def test_bytes(self):
        code = ReadWrite()
        self.assertEqual(b'[READ-WRITE]', bytes(code))


class TestTryCreate(unittest.TestCase):

    def test_bytes(self):
        code = TryCreate()
        self.assertEqual(b'[TRYCREATE]', bytes(code))


class TestUidNext(unittest.TestCase):

    def test_bytes(self):
        code = UidNext(13)
        self.assertEqual(b'[UIDNEXT 13]', bytes(code))


class TestUidValidity(unittest.TestCase):

    def test_bytes(self):
        code = UidValidity(1234)
        self.assertEqual(b'[UIDVALIDITY 1234]', bytes(code))


class TestUnseen(unittest.TestCase):

    def test_bytes(self):
        code = Unseen(3)
        self.assertEqual(b'[UNSEEN 3]', bytes(code))


class TestAppendUid(unittest.TestCase):

    def test_bytes(self):
        code = AppendUid(12345, [1, 2, 3, 5])
        self.assertEqual(b'[APPENDUID 12345 1:3,5]', bytes(code))


class TestCopyUid(unittest.TestCase):

    def test_bytes(self):
        code = CopyUid(12345, [(1, 100), (2, 101), (3, 102), (5, 103)])
        self.assertEqual(b'[COPYUID 12345 1:3,5 100:103]', bytes(code))




import errno
import os
import os.path
from abc import abstractmethod, ABCMeta
from mailbox import Maildir, NoSuchMailboxError  # type: ignore
from typing import Type, Sequence, Iterable
from typing_extensions import Protocol

__all__ = ['MaildirLayout', 'DefaultLayout', 'FilesystemLayout']

_Parts = Sequence[str]


class MaildirLayout(Protocol):
    """Manages the folder layout of a :class:`~mailbox.Maildir` inbox.

    See Also:
        `Directory Structure
        <https://wiki.dovecot.org/MailboxFormat/Maildir#line-130>`_

    """

    @classmethod
    def get(cls, path: str, layout: str,
            maildir_type: Type[Maildir] = Maildir) -> 'MaildirLayout':
        """

        Args:
            path: The root path of the inbox.
            layout: The layout name, e.g. ``'++'`` or ``'fs'``.
            maildir_type: The :class:`~mailbox.Maildir` class override.

        Raises:
            ValueError: The layout name was not recognized.

        """
        if layout == '++':
            return DefaultLayout(path, maildir_type)
        elif layout == 'fs':
            return FilesystemLayout(path, maildir_type)
        else:
            raise ValueError(layout)

    @property
    @abstractmethod
    def path(self) -> str:
        """The root path of the inbox."""
        ...

    @abstractmethod
    def get_path(self, name: str, delimiter: str) -> str:
        """Return the path of the sub-folder.

        Args:
            name: The nested parts of the folder name, not including inbox.
            delimiter: The nested sub-folder delimiter string.

        """
        ...

    @abstractmethod
    def list_folders(self, delimiter: str, top: str = 'INBOX') \
            -> Sequence[str]:
        """Return all folders, starting with ``top`` and traversing
        through the sub-folder heirarchy.

        Args:
            delimiter: The nested sub-folder delimiter string.
            top: The top of the folder heirarchy to list.

        Raises:
            FileNotFoundError: The folder did not exist.

        """
        ...

    @abstractmethod
    def get_folder(self, name: str, delimiter: str) -> Maildir:
        """Return the existing sub-folder.

        Args:
            name: The delimited sub-folder name, not including inbox.
            delimiter: The nested sub-folder delimiter string.

        Raises:
            FileNotFoundError: The folder did not exist.

        """
        ...

    @abstractmethod
    def add_folder(self, name: str, delimiter: str) -> Maildir:
        """Add and return a new sub-folder.

        Args:
            name: The delimited sub-folder name, not including inbox.
            delimiter: The nested sub-folder delimiter string.

        Raises:
            FileExistsError: The folder already exists.
            FileNotFoundError: A parent folder did not exist.

        """
        ...

    @abstractmethod
    def remove_folder(self, name: str, delimiter: str) -> None:
        """Remove the existing sub-folder.

        Args:
            name: The delimited sub-folder name, not including inbox.
            delimiter: The nested sub-folder delimiter string.

        Raises:
            FileNotFoundError: The folder did not exist.
            OSError: With :attr:`~errno.ENOTEMPTY`, the folder has sub-folders.

        """
        ...

    @abstractmethod
    def rename_folder(self, source_name: str, dest_name: str,
                      delimiter: str) -> Maildir:
        """Rename the existing sub-folder to the destination.

        Args:
            source_name: The delimited source sub-folder name.
            dest_name: The delimited destination sub-folder name.
            delimiter: The nested sub-folder delimiter string.

        Raises:
            FileNotFoundError: The source folder did not exist.
            FileExistsError: The destination folder already exists.

        """
        ...


class _BaseLayout(MaildirLayout, metaclass=ABCMeta):

    def __init__(self, path: str, maildir_type: Type[Maildir] = Maildir) \
            -> None:
        super().__init__()
        self._path = path
        self._maildir = maildir_type

    @property
    def path(self) -> str:
        return self._path

    @classmethod
    def _split(cls, name: str, delimiter: str) -> _Parts:
        if name == 'INBOX':
            return []
        return name.split(delimiter)

    @classmethod
    def _join(cls, parts: _Parts, delimiter: str) -> str:
        if not parts:
            return 'INBOX'
        return delimiter.join(parts)

    def _can_remove(self, parts: _Parts) -> bool:
        return True

    @abstractmethod
    def _get_path(self, parts: _Parts) -> str:
        ...

    @abstractmethod
    def _list_folders(self, parts: _Parts) -> Iterable[_Parts]:
        ...

    @abstractmethod
    def _rename_folder(self, source_parts: _Parts,
                       dest_parts: _Parts) -> Maildir:
        ...

    def get_path(self, name: str, delimiter: str) -> str:
        parts = self._split(name, delimiter)
        return self._get_path(parts)

    def list_folders(self, delimiter: str, top: str = 'INBOX') \
            -> Sequence[str]:
        parts = self._split(top, delimiter)
        return [self._join(sub_parts, delimiter)
                for sub_parts in self._list_folders(parts)]

    def get_folder(self, name: str, delimiter: str) -> Maildir:
        path = self.get_path(name, delimiter)
        try:
            return self._maildir(path, create=False)
        except NoSuchMailboxError:
            raise FileNotFoundError(path)

    def add_folder(self, name: str, delimiter: str) -> Maildir:
        parts = self._split(name, delimiter)
        for i in range(1, len(parts) - 1):
            path = self._get_path(parts[0:i])
            if not os.path.isdir(path):
                raise FileNotFoundError(path)
        path = self._get_path(parts)
        maildir = self._maildir(path, create=True)
        maildirfolder = os.path.join(path, 'maildirfolder')
        with open(maildirfolder, 'x'):
            pass
        return maildir

    def remove_folder(self, name: str, delimiter: str) -> None:
        parts = self._split(name, delimiter)
        path = self._get_path(parts)
        if not self._can_remove(parts):
            path = self._get_path(parts)
            raise OSError(errno.ENOTEMPTY, 'Directory not empty: '
                          + repr(path))
        for root, dirs, files in os.walk(path, topdown=False):
            for entry in files:
                os.remove(os.path.join(root, entry))
            for entry in dirs:
                os.rmdir(os.path.join(root, entry))
        os.rmdir(path)

    def rename_folder(self, source_name: str, dest_name: str,
                      delimiter: str) -> Maildir:
        source_parts = self._split(source_name, delimiter)
        dest_parts = self._split(dest_name, delimiter)
        for i in range(1, len(dest_parts) - 1):
            parts = dest_parts[0:i]
            path = self._get_path(parts)
            if not os.path.isdir(path):
                name = self._join(parts, delimiter)
                self.add_folder(name, delimiter)
        return self._rename_folder(source_parts, dest_parts)


class DefaultLayout(_BaseLayout):
    """The default Maildir++ layout, which uses sub-folder names starting
    with and using ``.`` as a delimiter for nesting , e.g.::

        .Trash/
        .Important.To-Do/
        .Important.Misc/

    Args:
        path: The root path of the inbox.
        maildir_type: The :class:`~mailbox.Maildir` class override.

    """

    def _get_path(self, parts: _Parts) -> str:
        return os.path.join(self._path, self._get_subdir(parts))

    @classmethod
    def _get_subdir(cls, parts: _Parts) -> str:
        if not parts:
            return ''
        return '.' + '.'.join(parts)

    @classmethod
    def _get_parts(cls, subdir: str) -> _Parts:
        if not subdir:
            return []
        return subdir[1:].split('.')

    def _list_folders(self, parts: _Parts) -> Iterable[_Parts]:
        subdir = self._get_subdir(parts)
        path = self._get_path(parts)
        if not os.path.isdir(path):
            return
        yield parts
        for elem in os.listdir(self._path):
            if elem in ('new', 'cur', 'tmp'):
                pass
            elif not subdir or elem.startswith(subdir + '.'):
                elem_path = os.path.join(self._path, elem)
                if os.path.isdir(elem_path):
                    yield self._get_parts(elem)

    def _rename_folder(self, source_parts: _Parts,
                       dest_parts: _Parts) -> Maildir:
        subdir = self._get_subdir(source_parts)
        dest_subdir = self._get_subdir(dest_parts)
        dest_path = self._get_path(dest_parts)
        for elem in os.listdir(self._path):
            if elem == subdir or elem.startswith(subdir + '.'):
                elem_path = os.path.join(self._path, elem)
                if os.path.isdir(elem_path):
                    dest_elem = dest_subdir + elem[len(subdir):]
                    dest_elem_path = os.path.join(self._path, dest_elem)
                    os.rename(elem_path, dest_elem_path)
        return self._maildir(dest_path, create=False)


class FilesystemLayout(_BaseLayout):
    """The ``fs`` layout, which uses nested sub-directories on the filesystem,
    e.g.::

        Trash/
        Important/To-Do/
        Important/Misc/

    Args:
        path: The root path of the inbox.
        maildir_type: The :class:`~mailbox.Maildir` class override.

    """

    def _get_path(self, parts: _Parts) -> str:
        return os.path.join(self._path, *parts)

    def _can_remove(self, parts: _Parts) -> bool:
        path = self._get_path(parts)
        for elem in os.listdir(path):
            if elem not in ('new', 'cur', 'tmp'):
                elem_path = os.path.join(path, elem)
                if os.path.isdir(elem_path):
                    return False
        return True

    def _list_folders(self, parts: _Parts) -> Iterable[_Parts]:
        path = self._get_path(parts)
        if not os.path.isdir(path):
            return
        yield parts
        for elem in os.listdir(path):
            if elem not in ('new', 'cur', 'tmp'):
                for sub_parts in self._list_folders(list(parts) + [elem]):
                    yield sub_parts

    def _rename_folder(self, source_parts: _Parts,
                       dest_parts: _Parts) -> Maildir:
        path = self._get_path(source_parts)
        dest_path = self._get_path(dest_parts)
        os.rename(path, dest_path)
        return self._maildir(dest_path, create=False)



from typing import Dict, List, Type, Tuple

from . import Space, Params
from .command import Command,  any, auth, nonauth, select
from .exceptions import NotParseable, CommandNotFound, CommandInvalid
from .primitives import Atom
from .specials import Tag

__all__ = ['Commands']


class Commands:
    """Contains the set of all known IMAP commands and the ability to parse."""

    __slots__ = ['commands']

    def __init__(self) -> None:
        super().__init__()
        self.commands: Dict[bytes, Type[Command]] = {}
        self._load_commands()

    def _load_commands(self):
        for mod in (any, auth, nonauth, select):
            for command_name in mod.__all__:
                command = getattr(mod, command_name)
                self.commands[command.command] = command

    def parse(self, buf: bytes, params: Params) -> Tuple[Command, bytes]:
        """Parse the given bytes into a command. The basic syntax is a tag
        string, a command name, possibly some arguments, and then an endline.

        Args:
            buf: The bytes to parse.
            params: The parsing parameters.

        """
        tag = Tag(b'*')
        try:
            tag, buf = Tag.parse(buf, params)
        except NotParseable:
            raise CommandNotFound(tag.value)
        cmd_parts: List[bytes] = []
        while True:
            try:
                _, buf = Space.parse(buf, params)
                atom, buf = Atom.parse(buf, params)
                cmd_parts.append(atom.value.upper())
            except NotParseable:
                raise CommandNotFound(tag.value)
            command = b' '.join(cmd_parts)
            cmd_type = self.commands.get(command)
            if not cmd_type:
                raise CommandNotFound(tag.value, command)
            elif not cmd_type.compound:
                break
        params = params.copy(tag=tag.value, command_name=command)
        try:
            return cmd_type.parse(buf, params)
        except NotParseable as exc:
            raise CommandInvalid(tag.value, command) from exc



import os.path
import ssl
from argparse import Namespace
from concurrent.futures import Executor
from ssl import SSLContext
from typing import Sequence, Any, Optional, Mapping, TypeVar, Type, Callable

from pysasl import SASLAuth, AuthenticationCredentials

from .concurrent import Event, ReadWriteLock
from .parsing import Params
from .parsing.commands import Commands

__all__ = ['IMAPConfig', 'ConfigT', 'ConfigT_contra']

#: Type variable with an upper bound of :class:`IMAPConfig`.
ConfigT = TypeVar('ConfigT', bound='IMAPConfig')

#: Contravariant type variable with an upper bound of :class:`IMAPConfig`.
ConfigT_contra = TypeVar('ConfigT_contra', bound='IMAPConfig',
                         contravariant=True)


class IMAPConfig:
    """Configurable settings that control how the IMAP server operates and what
    extensions it supports.

    Args:
        args: The command-line arguments.
        debug: If true, prints all socket activity to stdout.
        executor: If given, all backend operations will be run inside this
            executor object, e.g. a thread pool.
        ssl_context: SSL context that will be used for opportunistic TLS.
            Alternatively, you can pass extra arguments ``cert_file`` and
            ``key_file`` and an SSL context will be created.
        starttls_enabled: True if opportunistic TLS should be supported.
        reject_insecure_auth: True if authentication mechanisms that transmit
            credentials in cleartext should be rejected on non-encrypted
            transports.
        preauth_credentials: If given, clients will pre-authenticate on
            connection using these credentials.
        max_append_len: The maximum allowed length of the message body to an
            ``APPEND`` command.
        bad_command_limit: The number of consecutive commands received from
            the client with parsing errors before the client is disconnected.
        disable_idle: Disable the ``IDLE`` capability.
        extra: Additional keywords used for special circumstances.

    Attributes:
        args: The command-line arguments.

    """

    def __init__(self, args: Namespace,
                 debug: bool = False,
                 executor: Executor = None,
                 ssl_context: SSLContext = None,
                 starttls_enabled: bool = True,
                 reject_insecure_auth: bool = True,
                 preauth_credentials: AuthenticationCredentials = None,
                 max_append_len: Optional[int] = 1000000000,
                 bad_command_limit: Optional[int] = 5,
                 disable_idle: bool = False,
                 **extra: Any) -> None:
        super().__init__()
        self.args = args
        self._debug = debug
        self._executor = executor
        self._ssl_context = ssl_context
        self._starttls_enabled = starttls_enabled
        self._reject_insecure_auth = reject_insecure_auth
        self._preauth_credentials = preauth_credentials
        self._max_append_len = max_append_len
        self._bad_command_limit = bad_command_limit
        self._disable_idle = disable_idle
        self._extra = extra

    @classmethod
    def parse_args(cls, args: Namespace, **extra: Any) -> Mapping[str, Any]:
        """Given command-line arguments, return a dictionary of keywords that
        should be passed in to the :class:`IMAPConfig` (or sub-class)
        constructor. Sub-classes should override this method as needed.

        Args:
            args: The arguments parsed from the command-line.
            extra: Additional keywords used by sub-classes.

        """
        ret = {'debug': args.debug,
               'reject_insecure_auth': not args.insecure_login}
        for key, val in extra.items():
            if val is not None:
                ret[key] = val
        return ret

    @classmethod
    def from_args(cls: Type[ConfigT], args: Namespace,
                  **extra: Any) -> ConfigT:
        """Build and return a new :class:`IMAPConfig` using command-line
        arguments.

        Args:
            args: The arguments parsed from the command-line.
            extra: Additional keywords used by sub-classes.

        """
        params = cls.parse_args(args, **extra)
        return cls(args, **params)

    @property
    def debug(self) -> bool:
        return self._debug

    @property
    def executor(self) -> Optional[Executor]:
        return self._executor

    @property
    def new_rwlock(self) -> Callable[[], ReadWriteLock]:
        if self._executor is None:
            return ReadWriteLock.for_asyncio
        else:
            return ReadWriteLock.for_threading

    @property
    def new_event(self) -> Callable[[], Event]:
        if self._executor is None:
            return Event.for_asyncio
        else:
            return Event.for_threading

    @property
    def bad_command_limit(self) -> Optional[int]:
        return self._bad_command_limit

    @property
    def ssl_context(self) -> Optional[SSLContext]:
        if self._ssl_context is None:
            cert_file: Optional[str] = getattr(self.args, 'cert', None)
            if cert_file is None:
                return None
            key_file: str = getattr(self.args, 'key', cert_file)
            cert_path = os.path.realpath(cert_file)
            key_path = os.path.realpath(key_file)
            ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_context.load_cert_chain(cert_path, key_path)
            self._ssl_context = ssl_context
        return self._ssl_context

    @property
    def commands(self) -> Commands:
        return Commands()

    @property
    def initial_auth(self) -> SASLAuth:
        if self._reject_insecure_auth:
            return SASLAuth.secure()
        else:
            return SASLAuth()

    @property
    def starttls_auth(self) -> SASLAuth:
        return SASLAuth()

    @property
    def preauth_credentials(self) -> Optional[AuthenticationCredentials]:
        return self._preauth_credentials

    @property
    def parsing_params(self) -> Params:
        return Params(max_append_len=self._max_append_len)

    @property
    def login_capability(self) -> Sequence[bytes]:
        ret = [b'BINARY', b'UIDPLUS', b'MULTIAPPEND', b'CHILDREN']
        if not self._disable_idle:
            ret.append(b'IDLE')
        if self._max_append_len is not None:
            ret.append(b'APPENDLIMIT=%i' % self._max_append_len)
        return ret

    @property
    def initial_capability(self) -> Sequence[bytes]:
        ret = []
        if self._starttls_enabled:
            ret.append(b'STARTTLS')
        if self._reject_insecure_auth:
            ret.append(b'LOGINDISABLED')
        return ret




import unittest

from pymap.parsing import Params
from pymap.parsing.command import Command, CommandNoArgs
from pymap.parsing.commands import Commands
from pymap.parsing.exceptions import CommandNotFound, CommandInvalid, \
    NotParseable


class TestCommandInvalid(unittest.TestCase):

    def test_bytes(self):
        try:
            try:
                raise NotParseable(b'abc TEST stuff')
            except NotParseable as cause:
                raise CommandInvalid(b'abc', b'TEST') from cause
        except CommandInvalid as exc:
            self.assertEqual(b'abc', exc.tag)
            self.assertEqual(b'TEST', exc.command)
            self.assertEqual(b'TEST: [:ERROR:]abc TEST stuff', bytes(exc))
            self.assertEqual('TEST: [:ERROR:]abc TEST stuff', str(exc))


class TestCommandNotFound(unittest.TestCase):

    def test_bytes(self):
        exc = CommandNotFound(b'', b'TEST')
        self.assertEqual(b'Command Not Found: TEST', bytes(exc))

    def test_no_command(self):
        exc = CommandNotFound(b'')
        self.assertEqual(b'Command Not Given', bytes(exc))
        self.assertEqual('Command Not Given', str(exc))


class TestCommandNoArgs(unittest.TestCase):

    def test_parse(self):
        ret, buf = CommandNoArgs.parse(b'    \r\n test', Params())
        self.assertIsInstance(ret, CommandNoArgs)
        self.assertEqual(b' test', buf)

    def test_parse_no_cr(self):
        ret, buf = CommandNoArgs.parse(b'    \n test', Params())
        self.assertIsInstance(ret, CommandNoArgs)
        self.assertEqual(b' test', buf)


class TestCommands(unittest.TestCase):

    def setUp(self):
        self.commands = Commands()

    def test_parse(self):
        cmd, buf = self.commands.parse(b'a0 NOOP\n  ', Params())
        self.assertIsInstance(cmd, Command)
        self.assertEqual(b'a0', cmd.tag)
        self.assertEqual(b'  ', buf)




import re
from collections import OrderedDict
from typing import Dict, Iterable, Sequence, List, NamedTuple

__all__ = ['ListEntry', 'ListTree']


class ListEntry(NamedTuple):
    """An entry in the list results.

    Args:
        name: The name of the mailbox.
        exists: False if the mailbox should be marked ``\\Noselect``.
        has_children: Whether the mailbox should be marked ``\\HasChildren`` or
            ``\\HasNoChildren``.

    """

    name: str
    exists: bool
    has_children: bool

    @property
    def attributes(self) -> Sequence[bytes]:
        """The mailbox attributes that should be returned with the mailbox
        in a ``LIST`` response, e.g. ``\\Noselect``.

        See Also:
            `RFC 3348 <https://tools.ietf.org/html/rfc3348>`_

        """
        ret: List[bytes] = []
        if not self.exists:
            ret.append(b'Noselect')
        if self.has_children:
            ret.append(b'HasChildren')
        else:
            ret.append(b'HasNoChildren')
        return ret


class _TreeNode:

    __slots__ = ['parent', 'name', 'exists', 'children']

    def __init__(self, name: str, parent: '_TreeNode' = None) -> None:
        super().__init__()
        self.parent = parent
        self.name = name
        self.exists = False
        self.children: Dict[str, '_TreeNode'] = OrderedDict()

    def add(self, node_name: str, *extra: str) -> None:
        child = self.children.get(node_name)
        if not child:
            self.children[node_name] = child = _TreeNode(node_name, self)
        if not extra:
            child.exists = True
        else:
            child.add(*extra)

    def flatten(self, delimiter: str) -> str:
        node = self
        parts: List[str] = []
        while node.parent is not None:
            parts.append(node.name)
            node = node.parent
        return delimiter.join(reversed(parts))


class ListTree:
    """Constructs a tree of hierarchical mailbox names. If a mailbox name
    has superior names in the heirarchy that do not exist, they are added as
    "unreferenced".

    Args:
        delimiter: The string delimiter for nested mailbox parts.

    """

    _asterisk_escape = re.escape('*')
    _percent_escape = re.escape('%')
    _anything = '.*?'

    __slots__ = ['_delimiter', '_no_delimiter', '_root']

    def __init__(self, delimiter: str) -> None:
        super().__init__()
        self._delimiter = delimiter
        self._no_delimiter = '[^' + re.escape(delimiter) + ']*?'
        self._root = _TreeNode('')

    def update(self, *names: str) -> 'ListTree':
        """Add all the mailbox names to the tree, filling in any missing nodes.

        Args:
            names: The names of the mailboxes.

        """
        for name in names:
            parts = name.split(self._delimiter)
            self._root.add(*parts)
        return self

    def _iter(self, node: _TreeNode) -> Iterable[ListEntry]:
        if node.parent is not None:
            name = node.flatten(self._delimiter)
            yield ListEntry(name, node.exists, bool(node.children))
        for child in node.children.values():
            for entry in self._iter(child):
                yield entry

    def list(self) -> Iterable[ListEntry]:
        """Return all the entries in the list tree."""
        for entry in self._iter(self._root):
            yield entry

    def list_matching(self, ref_name: str, filter_: str) \
            -> Iterable[ListEntry]:
        """Return all the entries in the list tree that match the given query.

        Args:
            ref_name: Mailbox reference name.
            filter_: Mailbox name with possible wildcards.

        """
        re_interpreted = '^' + re.escape(ref_name + filter_) + '$'
        re_interpreted = re_interpreted.replace(
            self._asterisk_escape, self._anything)
        re_interpreted = re_interpreted.replace(
            self._percent_escape, self._no_delimiter)
        compiled_canonical = re.compile(re_interpreted)
        for entry in self.list():
            if entry.name == 'INBOX':
                if re.match(re_interpreted, 'INBOX', re.IGNORECASE):
                    yield entry
            elif compiled_canonical.match(entry.name):
                yield entry




import unittest
from datetime import datetime, timezone, timedelta

from pymap.parsing import Params
from pymap.parsing.exceptions import NotParseable, UnexpectedType, \
    InvalidContent
from pymap.parsing.specials import AString, Tag, Mailbox, DateTime, Flag, \
    StatusAttribute, SequenceSet, FetchAttribute, SearchKey, ExtensionOptions


class TestAString(unittest.TestCase):

    def test_parse(self):
        ret, buf = AString.parse(b'  a001[+]  ', Params())
        self.assertIsInstance(ret, AString)
        self.assertEqual(b'a001[+]', ret.value)
        self.assertEqual(b'  ', buf)

    def test_parse_quoted(self):
        ret, buf = AString.parse(br'  "a bc \" d"  ', Params())
        self.assertIsInstance(ret, AString)
        self.assertEqual(b'a bc " d', ret.value)
        self.assertEqual(b'  ', buf)

    def test_parse_literal(self):
        ret, buf = AString.parse(
            b'  {4}\r\n', Params(continuations=[b'abcd  ']))
        self.assertIsInstance(ret, AString)
        self.assertEqual(b'abcd', ret.value)
        self.assertEqual(b'  ', buf)

    def test_bytes(self):
        a1 = AString(b'abc123')
        self.assertEqual(b'abc123', bytes(a1))
        a2 = AString(b' a (b) c ')
        self.assertEqual(b'" a (b) c "', bytes(a2))
        self.assertEqual(b'" a (b) c "', bytes(a2))


class TestTag(unittest.TestCase):

    def test_parse(self):
        ret, buf = Tag.parse(b' a[001]  ', Params())
        self.assertIsInstance(ret, Tag)
        self.assertEqual(b'a[001]', ret.value)
        self.assertEqual(b'  ', buf)

    def test_parse_failure(self):
        with self.assertRaises(NotParseable):
            Tag.parse(b'', Params())

    def test_bytes(self):
        tag1 = Tag(b'a[001]')
        self.assertEqual(b'a[001]', bytes(tag1))


class TestMailbox(unittest.TestCase):

    def test_parse(self):
        ret, buf = Mailbox.parse(
            b'~peter/mail/&-/&U,BTFw-/&ZeVnLIqe-', Params())
        self.assertIsInstance(ret, Mailbox)
        self.assertEqual('~peter/mail/&/台北/日本語', ret.value)
        self.assertEqual(b'', buf)

    def test_parse_inbox(self):
        ret, buf = Mailbox.parse(b'  iNbOx  ', Params())
        self.assertIsInstance(ret, Mailbox)
        self.assertEqual('INBOX', ret.value)
        self.assertEqual(b'  ', buf)

    def test_parse_quoted(self):
        ret, buf = Mailbox.parse(b'  "test mailbox \\"stuff\\""  ', Params())
        self.assertIsInstance(ret, Mailbox)
        self.assertEqual('test mailbox "stuff"', ret.value)
        self.assertEqual(b'  ', buf)

    def test_parse_failure(self):
        with self.assertRaises(NotParseable):
            Mailbox.parse(b'  ', Params())

    def test_bytes(self):
        mbx = Mailbox('~peter/mail/&/台北/日本語')
        self.assertEqual(b'~peter/mail/&-/&U,BTFw-/&ZeVnLIqe-', bytes(mbx))
        self.assertEqual(b'~peter/mail/&-/&U,BTFw-/&ZeVnLIqe-', bytes(mbx))

    def test_str(self):
        mbx = Mailbox('~peter/mail/&/台北/日本語')
        self.assertEqual('~peter/mail/&/台北/日本語', str(mbx))


class TestDateTime(unittest.TestCase):

    def test_parse(self):
        ret, buf = DateTime.parse(b'"1-Jan-2000 01:02:03 +0500"', Params())
        self.assertIsInstance(ret, DateTime)
        self.assertEqual(1, ret.value.day)
        self.assertEqual(1, ret.value.month)
        self.assertEqual(2000, ret.value.year)
        self.assertEqual(1, ret.value.hour)
        self.assertEqual(2, ret.value.minute)
        self.assertEqual(3, ret.value.second)
        self.assertEqual(18000.0, ret.value.utcoffset().total_seconds())

    def test_parse_failure(self):
        with self.assertRaises(InvalidContent):
            DateTime.parse(b'"test"', Params())

    def test_bytes(self):
        dt1 = DateTime(datetime(2000, 1, 1, 1, 2, 3,
                                tzinfo=timezone(timedelta(hours=5))))
        self.assertEqual(b'"01-Jan-2000 01:02:03 +0500"', bytes(dt1))
        dt2 = DateTime(datetime.now(), b'testing')
        self.assertEqual(b'"testing"', bytes(dt2))


class TestFlag(unittest.TestCase):

    def test_parse(self):
        ret1, buf1 = Flag.parse(br'MyFlag  \Seen  ', Params())
        self.assertEqual(b'MyFlag', ret1.value)
        self.assertEqual(br'  \Seen  ', buf1)
        ret2, buf2 = Flag.parse(buf1, Params())
        self.assertEqual(br'\Seen', ret2.value)
        self.assertEqual(b'  ', buf2)

    def test_bytes(self):
        f1 = Flag(br'\testflag')
        self.assertEqual(br'\Testflag', bytes(f1))
        f2 = Flag(b'testflag')
        self.assertEqual(b'testflag', bytes(f2))


class TestStatusAttribute(unittest.TestCase):

    def test_valueerror(self):
        with self.assertRaises(ValueError):
            StatusAttribute(b'TEST')
        StatusAttribute(b'MESSAGES')

    def test_parse(self):
        ret, buf = StatusAttribute.parse(b'  messages  ', Params())
        self.assertEqual(b'MESSAGES', ret.value)
        self.assertEqual(b'  ', buf)

    def test_parse_invalid(self):
        with self.assertRaises(InvalidContent):
            StatusAttribute.parse(b'test', Params())

    def test_bytes(self):
        attr = StatusAttribute(b'MESSAGES')
        self.assertEqual(b'MESSAGES', bytes(attr))


class TestSequenceSet(unittest.TestCase):

    def test_parse(self):
        ret, buf = SequenceSet.parse(b'12,*,1:*,*:1  ', Params())
        self.assertEqual([12, '*', (1, '*'), ('*', 1)], ret.value)
        self.assertEqual(b'  ', buf)

    def test_parse_error(self):
        with self.assertRaises(NotParseable):
            SequenceSet.parse(b'*,test', Params())
        with self.assertRaises(NotParseable):
            SequenceSet.parse(b'*:test', Params())
        with self.assertRaises(NotParseable):
            SequenceSet.parse(b'', Params())
        with self.assertRaises(NotParseable):
            SequenceSet.parse(b'0:*', Params())


class TestFetchAttribute(unittest.TestCase):

    def test_hash(self):
        attr1 = FetchAttribute(b'TEST')
        attr2 = FetchAttribute(b'TEST')
        self.assertEqual(hash(attr1), hash(attr2))

    def test_parse(self):
        ret, buf = FetchAttribute.parse(
            b'body.peek[1.2.HEADER.FIELDS (A B)]<4.5>  ', Params())
        self.assertEqual(b'BODY.PEEK', ret.value)
        self.assertEqual([1, 2], ret.section.parts)
        self.assertEqual(b'HEADER.FIELDS', ret.section.specifier)
        self.assertEqual({'A', 'B'}, ret.section.headers)
        self.assertEqual((4, 5), ret.partial)
        self.assertEqual(b'  ', buf)

    def test_parse_simple(self):
        ret1, _ = FetchAttribute.parse(b'ENVELOPE', Params())
        self.assertEqual(b'ENVELOPE', ret1.value)
        self.assertIsNone(ret1.section)
        self.assertIsNone(ret1.partial)
        ret2, _ = FetchAttribute.parse(b'BODY', Params())
        self.assertEqual(b'BODY', ret2.value)
        self.assertIsNone(ret2.section)
        self.assertIsNone(ret2.partial)

    def test_parse_sections(self):
        ret1, _ = FetchAttribute.parse(b'BODY[1.2]', Params())
        self.assertEqual(b'BODY', ret1.value)
        self.assertEqual([1, 2], ret1.section.parts)
        self.assertIsNone(ret1.section.specifier)
        self.assertIsNone(ret1.section.headers)
        self.assertIsNone(ret1.partial)
        ret2, _ = FetchAttribute.parse(b'BODY[1.2.MIME]', Params())
        self.assertEqual(b'BODY', ret2.value)
        self.assertEqual([1, 2], ret2.section.parts)
        self.assertEqual(b'MIME', ret2.section.specifier)
        self.assertIsNone(ret2.section.headers)
        self.assertIsNone(ret2.partial)
        ret3, _ = FetchAttribute.parse(b'BODY[HEADER]', Params())
        self.assertEqual(b'BODY', ret3.value)
        self.assertEqual([], ret3.section.parts)
        self.assertEqual(b'HEADER', ret3.section.specifier)
        self.assertIsNone(ret3.section.headers)
        self.assertIsNone(ret3.partial)

    def test_parse_error(self):
        with self.assertRaises(NotParseable):
            FetchAttribute.parse(b'<>', Params())
        with self.assertRaises(NotParseable):
            FetchAttribute.parse(b'BODY[""]', Params())
        with self.assertRaises(NotParseable):
            FetchAttribute.parse(b'BODY[HEADER.FIELDS ()]', Params())
        with self.assertRaises(NotParseable):
            FetchAttribute.parse(b'BODY[TEST]', Params())
        with self.assertRaises(NotParseable):
            FetchAttribute.parse(b'TEST', Params())
        with self.assertRaises(NotParseable):
            FetchAttribute.parse(b'BODY.PEEK', Params())
        with self.assertRaises(NotParseable):
            FetchAttribute.parse(b'BODY[TEXT]<10.0>', Params())

    def test_bytes(self):
        attr1 = FetchAttribute(b'ENVELOPE')
        self.assertEqual(b'ENVELOPE', bytes(attr1))
        self.assertEqual(b'ENVELOPE', bytes(attr1))
        section = FetchAttribute.Section((1, 2), b'STUFF',
                                         frozenset({'A', 'B'}))
        attr2 = FetchAttribute(b'BODY', section, (4, 5))
        self.assertEqual(b'BODY[1.2.STUFF (A B)]<4.5>', bytes(attr2))
        self.assertEqual(b'BODY[1.2.STUFF (A B)]<4>',
                         bytes(attr2.for_response))


class TestSearchKey(unittest.TestCase):

    def test_parse(self):
        ret, buf = SearchKey.parse(b'ALL  ', Params())
        self.assertEqual(b'ALL', ret.value)
        self.assertIsNone(ret.filter)
        self.assertFalse(ret.inverse)
        self.assertEqual(b'  ', buf)

    def test_parse_seqset(self):
        ret, buf = SearchKey.parse(b'NOT 1,2,3', Params())
        self.assertEqual(b'SEQSET', ret.value)
        self.assertIsInstance(ret.filter, SequenceSet)
        self.assertEqual([1, 2, 3], ret.filter.value)
        self.assertTrue(ret.inverse)

    def test_parse_list(self):
        ret, buf = SearchKey.parse(b'(4,5,6 NOT 1,2,3)', Params())
        self.assertEqual(b'KEYSET', ret.value)
        self.assertIsInstance(ret.filter, list)
        self.assertEqual(2, len(ret.filter))
        self.assertEqual(b'SEQSET', ret.filter[0].value)
        self.assertIsInstance(ret.filter[0].filter, SequenceSet)
        self.assertEqual([4, 5, 6],
                         ret.filter[0].filter.value)
        self.assertFalse(ret.filter[0].inverse)
        self.assertEqual(b'SEQSET', ret.filter[1].value)
        self.assertIsInstance(ret.filter[1].filter, SequenceSet)
        self.assertEqual([1, 2, 3],
                         ret.filter[1].filter.value)
        self.assertTrue(ret.filter[1].inverse)

    def test_parse_list_error(self):
        with self.assertRaises(UnexpectedType):
            SearchKey.parse(b'(TEST)', Params())
        with self.assertRaises(NotParseable):
            SearchKey.parse(b'(1,2,3', Params())

    def test_parse_filter_astring(self):
        ret, buf = SearchKey.parse(b'subject "testing"', Params())
        self.assertEqual(b'SUBJECT', ret.value)
        self.assertEqual('testing', ret.filter)
        self.assertFalse(ret.inverse)

    def test_parse_filter_date(self):
        ret, buf = SearchKey.parse(b'before 01-Jan-1970', Params())
        self.assertEqual(b'BEFORE', ret.value)
        self.assertEqual(datetime(1970, 1, 1), ret.filter)
        self.assertFalse(ret.inverse)

    def test_parse_filter_keyword(self):
        ret, buf = SearchKey.parse(b'keyword test', Params())
        self.assertEqual(b'KEYWORD', ret.value)
        self.assertEqual(b'test', ret.filter)
        self.assertFalse(ret.inverse)

    def test_parse_filter_number(self):
        ret, buf = SearchKey.parse(b'larger 1000', Params())
        self.assertEqual(b'LARGER', ret.value)
        self.assertEqual(1000, ret.filter)
        self.assertFalse(ret.inverse)

    def test_parse_filter_uid(self):
        ret, buf = SearchKey.parse(b'UID 1,2,3', Params())
        self.assertEqual(b'SEQSET', ret.value)
        self.assertIsInstance(ret.filter, SequenceSet)
        self.assertEqual([1, 2, 3], ret.filter.value)
        self.assertFalse(ret.inverse)

    def test_parse_filter_header(self):
        ret, buf = SearchKey.parse(
            b'HEADER "From" "test@example.com"', Params())
        self.assertEqual(b'HEADER', ret.value)
        self.assertEqual(('From', 'test@example.com'), ret.filter)
        self.assertFalse(ret.inverse)

    def test_parse_filter_or(self):
        ret, buf = SearchKey.parse(b'OR NEW OLD', Params())
        self.assertEqual(b'OR', ret.value)
        self.assertEqual(b'NEW', ret.filter[0].value)
        self.assertEqual(b'OLD', ret.filter[1].value)
        self.assertFalse(ret.inverse)

    def test_parse_error(self):
        with self.assertRaises(NotParseable):
            SearchKey.parse(b'test', Params())
        with self.assertRaises(NotParseable):
            SearchKey.parse(b'before stuff', Params())


class TestExtensionOptions(unittest.TestCase):

    def test_parse_empty(self):
        ret, buf = ExtensionOptions.parse(b'()', Params())
        self.assertEqual({}, ret.value)
        self.assertEqual(b'', buf)
        ret, buf = ExtensionOptions.parse(b'', Params())
        self.assertEqual({}, ret.value)
        self.assertEqual(b'', buf)

    def test_parse(self):
        ret, buf = ExtensionOptions.parse(b'(one two)abc', Params())
        self.assertEqual({b'ONE': [], b'TWO': []}, ret.value)
        self.assertEqual(b'abc', buf)

    def test_parse_num(self):
        ret, buf = ExtensionOptions.parse(b'(one 1 two 2)abc', Params())
        self.assertEqual({b'ONE': [b'1'], b'TWO': [b'2']}, ret.value)
        self.assertEqual(b'abc', buf)

    def test_parse_seqset(self):
        ret, buf = ExtensionOptions.parse(b'(one 1:3 two 2)abc', Params())
        self.assertEqual({b'ONE': [b'1:3'], b'TWO': [b'2']}, ret.value)
        self.assertEqual(b'abc', buf)

    def test_parse_string(self):
        ret, buf = ExtensionOptions.parse(b'(one (two) three)abc', Params())
        self.assertEqual({b'ONE': [b'two'], b'THREE': []}, ret.value)
        self.assertEqual(b'abc', buf)

    def test_parse_nested(self):
        ret, buf = ExtensionOptions.parse(b'(one (two (three)))abc', Params())
        self.assertEqual({b'ONE': [b'two', [b'three']]}, ret.value)
        self.assertEqual(b'abc', buf)

    def test_bytes(self):
        ret, _ = ExtensionOptions.parse(b'()', Params())
        self.assertEqual(b'()', bytes(ret))
        ret, _ = ExtensionOptions.parse(b'(one two)', Params())
        self.assertEqual(b'(ONE TWO)', bytes(ret))
        ret, _ = ExtensionOptions.parse(b'(one 1 two 2)', Params())
        self.assertEqual(b'(ONE 1 TWO 2)', bytes(ret))
        ret, _ = ExtensionOptions.parse(b'(one 1:3 two 2)', Params())
        self.assertEqual(b'(ONE 1:3 TWO 2)', bytes(ret))
        ret, _ = ExtensionOptions.parse(b'(one (two) three)', Params())
        self.assertEqual(b'(ONE (two) THREE)', bytes(ret))
        ret, _ = ExtensionOptions.parse(b'(one (two (three)))', Params())
        self.assertEqual(b'(ONE (two (three)))', bytes(ret))



"""Contains the project's version number in ``__version__``."""

import pkg_resources

__all__ = ['__version__']

#: The project version string.
__version__: str = pkg_resources.require("pymap")[0].version



"""Package defining all the IMAP parsing and response classes."""

from typing import Optional

from .response import ResponseBad, ResponseCode
from ..bytes import MaybeBytes

__all__ = ['RequiresContinuation', 'NotParseable', 'InvalidContent',
           'UnexpectedType', 'BadCommand', 'CommandNotFound', 'CommandInvalid']


class RequiresContinuation(Exception):
    """Indicates that the buffer has been successfully parsed so far, but
    requires a continuation of the command from the client.

    Args:
        message: The message from the server.
        literal_length: If the continuation is for a string literal, this is
            the byte length to expect.

    """

    __slots__ = ['message', 'literal_length']

    def __init__(self, message: bytes, literal_length: int = 0) -> None:
        super().__init__()
        self.message = message
        self.literal_length = literal_length


class NotParseable(Exception):
    """Indicates that the given buffer was not parseable by one or all of the
    data formats.

    Args:
        buf: The buffer with the parsing error.

    """

    error_indicator = b'[:ERROR:]'

    __slots__ = ['buf', 'code', 'offset', '_raw', '_before', '_after']

    def __init__(self, buf: bytes, code: Optional[MaybeBytes] = None) -> None:
        super().__init__()
        self.buf = buf
        self.code = ResponseCode.of(code)
        self.offset: int = 0
        self._raw: Optional[bytes] = None
        self._before: Optional[bytes] = None
        self._after: Optional[bytes] = None
        if isinstance(buf, memoryview):
            obj = getattr(buf, 'obj')
            nbytes = getattr(buf, 'nbytes')
            self.offset = len(obj) - nbytes

    @property
    def before(self) -> bytes:
        """The bytes before the parsing error was encountered."""
        if self._before is not None:
            return self._before
        buf = self.buf
        if isinstance(buf, memoryview):
            buf = getattr(buf, 'obj')
        self._before = before = buf[0:self.offset]
        return before

    @property
    def after(self) -> bytes:
        """The bytes after the parsing error was encountered."""
        if self._after is not None:
            return self._after
        if isinstance(self.buf, memoryview):
            self._after = after = self.buf.tobytes()
        else:
            self._after = after = self.buf
        return after

    def __bytes__(self) -> bytes:
        if self._raw is not None:
            return self._raw
        before = self.before
        after = self.after.rstrip(b'\r\n')
        self._raw = raw = self.error_indicator.join((before, after))
        return raw

    def __str__(self) -> str:
        return str(bytes(self), 'ascii', 'replace')


class InvalidContent(NotParseable, ValueError):
    """Indicates the type of the parsed content was correct, but something
    about the content did not fit what was expected by the special type.

    """
    pass


class UnexpectedType(NotParseable):
    """Indicates that a generic parseable that was given a sub-type expectation
    failed to meet that expectation.

    """
    pass


class BadCommand(Exception):
    """Base class for errors that occurred while parsing a command received
    from the client.

    Args:
        tag: The command tag value.

    """

    __slots__ = ['tag']

    def __init__(self, tag: bytes) -> None:
        super().__init__()
        self.tag = tag

    @property
    def code(self) -> Optional[ResponseCode]:
        """The optional response code."""
        return None

    def __bytes__(self) -> bytes:
        raise NotImplementedError

    def __str__(self) -> str:
        return bytes(self).decode('ascii', 'ignore')

    def get_response(self) -> ResponseBad:
        """The response to send back to the client in response to the command
        parsing error.

        """
        return ResponseBad(self.tag, bytes(self), self.code)


class CommandNotFound(BadCommand):
    """Error indicating the data was not parseable because the command was not
    found.

    Args:
        tag: The command tag value.
        command: The command name.

    """

    __slots__ = ['command']

    def __init__(self, tag: bytes, command: bytes = None) -> None:
        super().__init__(tag)
        self.command = command

    def __bytes__(self) -> bytes:
        if self.command:
            return b'Command Not Found: %b' % self.command
        else:
            return b'Command Not Given'


class CommandInvalid(BadCommand):
    """Error indicating the data was not parseable because the command had
    invalid arguments.

    Args:
        tag: The command tag value.
        command: The command class.

    """

    def __init__(self, tag: bytes, command: bytes) -> None:
        super().__init__(tag)
        self.command = command
        self._raw: Optional[bytes] = None

    @property
    def code(self) -> Optional[ResponseCode]:
        return self.cause.code

    @property
    def cause(self) -> NotParseable:
        """The parsing exception that caused this exception."""
        exc = self.__cause__
        if not exc or not isinstance(exc, NotParseable):
            raise TypeError('Exception must have cause')
        return exc

    def __bytes__(self) -> bytes:
        if self._raw is None:
            self._raw = b': '.join((self.command, bytes(self.cause)))
        return self._raw




import asyncio
import threading
from contextvars import copy_context
from concurrent.futures import Executor
from functools import partial
from typing import cast, Optional

from .interfaces.session import LoginProtocol, SessionInterface

__all__ = ['ExecutorProxy']


class ExecutorProxy:
    """Proxies calls to :class:`~pymap.interfaces.session.SessionInterface`
    object or :class:`~pymap.interfaces.session.LoginProtocol`
    functions through an :class:`~concurrent.futures.Executor`.

    Args:
        executor: The executor to proxy through, e.g. a thread pool, or None.

    """

    def __init__(self, executor: Optional[Executor]) -> None:
        super().__init__()
        self.executor = executor
        self._local = _EventLoopLocal() if executor else None

    def wrap_login(self, login: LoginProtocol) -> LoginProtocol:
        """Wrap a login function so that it is called inside the executor."""
        if self.executor is None:
            return login
        else:
            wrapped = partial(self._call, login)
            return cast(LoginProtocol, wrapped)

    def wrap_session(self, session: SessionInterface) -> SessionInterface:
        """Wrap a session object so that all its methods are called inside the
        executor.

        """
        if self.executor is None:
            return session
        else:
            wrapped = _ProxyWrapper(self._call, session)
            return cast(SessionInterface, wrapped)

    async def _call(self, func, *args, **kwargs):
        loop = asyncio.get_event_loop()
        ctx = copy_context()
        func_call = partial(func, *args, **kwargs)
        return await loop.run_in_executor(
            self.executor, self._call_with_event_loop, func_call(), ctx)

    def _call_with_event_loop(self, future, ctx):
        loop = self._local.event_loop
        return ctx.run(loop.run_until_complete, future)


class _EventLoopLocal(threading.local):

    def __init__(self) -> None:
        self.event_loop = asyncio.new_event_loop()


class _ProxyWrapper:

    __slots__ = ['_call', '_obj']

    def __init__(self, call, obj) -> None:
        super().__init__()
        self._call = call
        self._obj = obj

    def __getattr__(self, key: str):
        func = getattr(self._obj, key)
        return partial(self._call, func)



import heapq
import math
import re
from itertools import chain
from typing import Iterable, Tuple, Union, Sequence, Optional, List

from .. import Params, Parseable, Space
from ..exceptions import NotParseable

__all__ = ['SequenceSet']

_SeqIdx = Union[None, str, int]
_SeqElem = Union[_SeqIdx, Tuple[_SeqIdx, _SeqIdx]]


class SequenceSet(Parseable[Sequence[_SeqElem]]):
    """Represents a sequence set from an IMAP stream.

    Args:
        sequences: The sequence set data.
        uid: True if the sequences refer to message UIDs.

    Attributes:
        uid: True if the sequences refer to message UIDs.

    """

    _num_pattern = re.compile(br'[1-9]\d*')

    def __init__(self, sequences: Sequence[_SeqElem],
                 uid: bool = False) -> None:
        super().__init__()
        self.sequences = sequences
        self.uid = uid
        self._flattened_cache = None
        self._raw: Optional[bytes] = None

    @classmethod
    def all(cls, uid: bool = False) -> 'SequenceSet':
        """A sequence set intended to contain all values."""
        return SequenceSet([(1, '*')], uid)

    @property
    def value(self) -> Sequence[_SeqElem]:
        """The sequence set data."""
        return self.sequences

    @property
    def _flattened(self):
        if self._flattened_cache is None:
            results = []
            for group in self.value:
                if isinstance(group, tuple):
                    group_left, group_right = group
                    if group_left == '*':
                        group_left = math.inf
                    if group_right == '*':
                        group_right = math.inf
                    high = max(group_left, group_right)
                    low = min(group_left, group_right)
                    heapq.heappush(results, (low, high))
                elif group == '*':
                    heapq.heappush(results, (math.inf, math.inf))
                else:
                    heapq.heappush(results, (group, group))
            flattened = []
            while results:
                lowest_min, lowest_max = heapq.heappop(results)
                while results:
                    next_lowest_min, next_lowest_max = results[0]
                    if lowest_max + 1 >= next_lowest_min:
                        heapq.heappop(results)
                        lowest_max = next_lowest_max
                    else:
                        break
                flattened.append((lowest_min, lowest_max))
            self._flattened_cache = flattened
        return self._flattened_cache

    def _get_flattened_bounded(self, max_value: int):
        for low, high in self._flattened:
            low = max_value if math.isinf(low) else low
            high = max_value if math.isinf(high) else high
            if low > max_value:
                break
            elif high > max_value:
                yield (low, max_value)
                break
            else:
                yield (low, high)

    def contains(self, num: int, max_value: int) -> bool:
        """Check if the sequence set contains the given value, when bounded
        by the given maximum value (in place of any ``'*'``).

        Args:
            num: The number to check.
            max_value: The maximum value of the set.

        """
        for low, high in self._get_flattened_bounded(max_value):
            if num < low:
                break
            elif num <= high:
                return True
        return False

    def iter(self, max_value: int) -> Iterable:
        """Iterates through the sequence numbers contained in the set, bounded
        by the given maximum value (in place of any ``'*'``).

        Args:
            max_value: The maximum value of the set.

        """
        return chain.from_iterable(
            [range(min(low, max_value), min(high, max_value) + 1)
             for low, high in self._get_flattened_bounded(max_value)])

    def __bytes__(self) -> bytes:
        if self._raw is not None:
            return self._raw
        parts = []
        for group in self.value:
            if isinstance(group, tuple):
                left = bytes(str(group[0]), 'ascii')
                right = bytes(str(group[1]), 'ascii')
                parts.append(b'%b:%b' % (left, right))
            else:
                parts.append(bytes(str(group), 'ascii'))
        self._raw = raw = b','.join(parts)
        return raw

    def __eq__(self, other) -> bool:
        if isinstance(other, SequenceSet):
            return self.uid == other.uid \
                and self._flattened == other._flattened
        return super().__eq__(other)

    def __hash__(self) -> int:
        return hash((tuple(self.sequences), self.uid))

    def __repr__(self) -> str:
        return '<%s set=%r>' % (type(self).__name__, self.sequences)

    @classmethod
    def _parse_part(cls, buf: bytes) -> Tuple[_SeqElem, bytes]:
        if buf and buf[0] == 0x2a:
            item1: _SeqIdx = '*'
            buf = buf[1:]
        else:
            match = cls._num_pattern.match(buf)
            if match:
                buf = buf[match.end(0):]
                item1 = int(match.group(0))
            else:
                raise NotParseable(buf)
        if buf and buf[0] == 0x3a:
            buf = buf[1:]
            if buf and buf[0] == 0x2a:
                return (item1, '*'), buf[1:]
            match = cls._num_pattern.match(buf)
            if match:
                buf = buf[match.end(0):]
                return (item1, int(match.group(0))), buf
            raise NotParseable(buf)
        return item1, buf

    @classmethod
    def build(cls, seqs: Iterable[int], uid: bool = False) -> 'SequenceSet':
        """Build a new sequence set that contains the given values using as
        few groups as possible.

        Args:
            seqs: The sequence values to build.
            uid: True if the sequences refer to message UIDs.

        """
        seqs_list = sorted(set(seqs))
        groups: List[Union[int, Tuple[int, int]]] = []
        group: Union[int, Tuple[int, int]] = seqs_list[0]
        for i in range(1, len(seqs_list)):
            group_i = seqs_list[i]
            if isinstance(group, int):
                if group_i == group + 1:
                    group = (group, group_i)
                else:
                    groups.append(group)
                    group = group_i
            elif isinstance(group, tuple):
                if group_i == group[1] + 1:
                    group = (group[0], group_i)
                else:
                    groups.append(group)
                    group = group_i
        groups.append(group)
        return SequenceSet(groups, uid)

    @classmethod
    def parse(cls, buf: bytes, params: Params) -> Tuple['SequenceSet', bytes]:
        try:
            _, buf = Space.parse(buf, params)
        except NotParseable:
            pass
        sequences = []
        while buf:
            item, buf = cls._parse_part(buf)
            sequences.append(item)
            if buf and buf[0] != 0x2c:
                break
            buf = buf[1:]
        if not sequences:
            raise NotParseable(buf)
        return cls(sequences, uid=params.uid), buf




from email.headerregistry import Address, AddressHeader, UnstructuredHeader, \
    DateHeader, ContentDispositionHeader
from itertools import chain
from typing import SupportsBytes, Optional, Dict, Sequence, Union

from ..primitives import ListP, Nil, Number, String
from ..specials import DateTime

__all__ = ['EnvelopeStructure', 'BodyStructure', 'MultipartBodyStructure',
           'ContentBodyStructure', 'TextBodyStructure', 'MessageBodyStructure']


class _Concatenated:

    def __init__(self, parts: Sequence[SupportsBytes]) -> None:
        super().__init__()
        self.parts = parts

    def __bytes__(self) -> bytes:
        return b''.join([bytes(part) for part in self.parts])


class _AddressList:

    def __init__(self, headers: Optional[Sequence[AddressHeader]]) -> None:
        super().__init__()
        self.headers: Sequence[AddressHeader] = headers or []

    @classmethod
    def _parse(cls, address: Address):
        realname = String.build(address.display_name)
        localpart = String.build(address.username)
        domain = String.build(address.domain)
        return ListP([realname, Nil(), localpart, domain])

    @property
    def _value(self) -> SupportsBytes:
        if self.headers:
            return ListP([self._parse(address)
                          for header in self.headers
                          for address in header.addresses])
        else:
            return Nil()

    def __bytes__(self) -> bytes:
        return bytes(self._value)


class _ParamsList:

    def __init__(self, params: Optional[Dict[str, str]]) -> None:
        super().__init__()
        self.params = params

    @property
    def _value(self) -> SupportsBytes:
        if self.params:
            values = [(String.build(key), String.build(value))
                      for key, value in self.params.items()]
            return ListP(chain.from_iterable(values))
        else:
            return Nil()

    def __bytes__(self) -> bytes:
        return bytes(self._value)


class EnvelopeStructure:
    """Builds the response to an `RFC 3501 7.4.2
    <https://tools.ietf.org/html/rfc3501#section-7.4.2>`_ FETCH ENVELOPE
    request.

    Args:
        date: Original date of the message.
        subject: ``Subject:`` header.
        from_: ``From:`` headers.
        sender: ``Sender:`` headers.
        reply_to: ``Reply-To:`` headers.
        to: ``To:`` headers.
        cc: ``Cc:`` headers.
        bcc: ``Bcc`` headers.
        in_reply_to: ``In-Reply-To:`` header.
        message_id: ``Message-Id:`` header.

    """

    def __init__(self, date: Optional[DateHeader],
                 subject: Optional[UnstructuredHeader],
                 from_: Optional[Sequence[AddressHeader]],
                 sender: Optional[Sequence[AddressHeader]],
                 reply_to: Optional[Sequence[AddressHeader]],
                 to: Optional[Sequence[AddressHeader]],
                 cc: Optional[Sequence[AddressHeader]],
                 bcc: Optional[Sequence[AddressHeader]],
                 in_reply_to: Optional[UnstructuredHeader],
                 message_id: Optional[UnstructuredHeader]) -> None:
        super().__init__()
        self.date = date
        self.subject = subject
        self.from_ = from_
        self.sender = sender
        self.reply_to = reply_to
        self.to = to
        self.cc = cc
        self.bcc = bcc
        self.in_reply_to = in_reply_to
        self.message_id = message_id

    def _addresses(self, headers: Optional[Sequence[AddressHeader]],
                   fallback: Optional[Sequence[AddressHeader]] = None) \
            -> SupportsBytes:
        if not headers and fallback:
            return self._addresses(fallback)
        return _AddressList(headers)

    @property
    def _value(self) -> SupportsBytes:
        datetime: Union[DateTime, Nil] = \
            DateTime(self.date.datetime) if self.date else Nil()
        return ListP([datetime,
                      String.build(self.subject),
                      self._addresses(self.from_),
                      self._addresses(self.sender, self.from_),
                      self._addresses(self.reply_to, self.from_),
                      self._addresses(self.to),
                      self._addresses(self.cc),
                      self._addresses(self.bcc),
                      String.build(self.in_reply_to),
                      String.build(self.message_id)])

    def __bytes__(self) -> bytes:
        return bytes(self._value)


class BodyStructure:
    """Parent class for the response to an `RFC 3501 7.4.2
    <https://tools.ietf.org/html/rfc3501#section-7.4.2>`_ FETCH BODYSTRUCTURE
    request. This class should not be used directly.

    Args:
        maintype: Main-type of the ``Content-Type:`` header.
        subtype: Sub-type of the ``Content-Type:`` header.
        content_type_params: Parameters from the ``Content-Type:`` header.
        content_disposition: ``Content-Disposition:`` header.
        content_language: ``Content-Language:`` header.
        content_location: ``Content-Location:`` header.

    """

    def __init__(self, maintype: str, subtype: str,
                 content_type_params: Optional[Dict[str, str]],
                 content_disposition: Optional[ContentDispositionHeader],
                 content_language: Optional[UnstructuredHeader],
                 content_location: Optional[UnstructuredHeader]) -> None:
        super().__init__()
        self.maintype = maintype
        self.subtype = subtype
        self.content_type_params = content_type_params
        self.content_disposition = content_disposition
        self.content_language = content_language
        self.content_location = content_location

    @property
    def _value(self) -> ListP:
        raise NotImplementedError

    def __bytes__(self) -> bytes:
        return bytes(self._value)

    @property
    def extended(self) -> ListP:
        """The body structure attributes with extension data."""
        raise NotImplementedError


class MultipartBodyStructure(BodyStructure):
    """:class:`BodyStructure` sub-class for ``multipart/*`` messages. The
    response is made up of the BODYSTRUCTUREs of all sub-parts.

    Args:
        subtype: Sub-type of the ``Content-Type:`` header.
        content_type_params: Parameters from the ``Content-Type:`` header.
        content_disposition: ``Content-Disposition:`` header.
        content_language: ``Content-Language:`` header.
        content_location: ``Content-Location:`` header.
        parts: Sub-part body structure objects.

    """

    def __init__(self, subtype: str,
                 content_type_params: Optional[Dict[str, str]],
                 content_disposition: Optional[ContentDispositionHeader],
                 content_language: Optional[UnstructuredHeader],
                 content_location: Optional[UnstructuredHeader],
                 parts: Sequence[BodyStructure]) -> None:
        super().__init__('multipart', subtype, content_type_params,
                         content_disposition, content_language,
                         content_location)
        self.parts = parts

    @property
    def _value(self) -> ListP:
        return ListP([_Concatenated(self.parts), String.build(self.subtype)])

    @property
    def extended(self) -> ListP:
        """The body structure attributes with extension data."""
        parts = [part.extended for part in self.parts]
        return ListP([_Concatenated(parts), String.build(self.subtype),
                      _ParamsList(self.content_type_params),
                      String.build(self.content_disposition),
                      String.build(self.content_language),
                      String.build(self.content_location)])


class ContentBodyStructure(BodyStructure):
    """:class:`BodyStructure` sub-class for any content type that does not
    fit the other sub-classes.

    Args:
        maintype: Main-type of the ``Content-Type:`` header.
        subtype: Sub-type of the ``Content-Type:`` header.
        content_type_params: Parameters from the ``Content-Type:`` header.
        content_disposition: ``Content-Disposition:`` header.
        content_language: ``Content-Language:`` header.
        content_location: ``Content-Location:`` header.
        content_id: ``Content-Id:`` header.
        content_description: ``Content-Description:`` header.
        content_transfer_encoding: ``Content-Transfer-Encoding:`` header.
        body_md5: MD5 hash of the body content.
        size: Size of the message body, in bytes.

    """

    def __init__(self, maintype: str, subtype: str,
                 content_type_params: Optional[Dict[str, str]],
                 content_disposition: Optional[ContentDispositionHeader],
                 content_language: Optional[UnstructuredHeader],
                 content_location: Optional[UnstructuredHeader],
                 content_id: Optional[UnstructuredHeader],
                 content_description: Optional[UnstructuredHeader],
                 content_transfer_encoding: Optional[UnstructuredHeader],
                 body_md5: Optional[str],
                 size: int) -> None:
        super().__init__(maintype, subtype, content_type_params,
                         content_disposition, content_language,
                         content_location)
        self.content_id = content_id
        self.content_description = content_description
        self.content_transfer_encoding = content_transfer_encoding
        self.body_md5 = body_md5
        self.size = size

    @property
    def _value(self) -> ListP:
        return ListP([String.build(self.maintype), String.build(self.subtype),
                      _ParamsList(self.content_type_params),
                      String.build(self.content_id),
                      String.build(self.content_description),
                      String.build(self.content_transfer_encoding),
                      Number(self.size)])

    @property
    def extended(self) -> ListP:
        """The body structure attributes with extension data."""
        return ListP([String.build(self.maintype), String.build(self.subtype),
                      _ParamsList(self.content_type_params),
                      String.build(self.content_id),
                      String.build(self.content_description),
                      String.build(self.content_transfer_encoding),
                      Number(self.size),
                      String.build(self.body_md5),
                      String.build(self.content_disposition),
                      String.build(self.content_language),
                      String.build(self.content_location)])


class TextBodyStructure(ContentBodyStructure):
    """:class:`BodyStructure` sub-class for ``text/*`` messages.

    Args:
        subtype: Sub-type of the ``Content-Type:`` header.
        content_type_params: Parameters from the ``Content-Type:`` header.
        content_disposition: ``Content-Disposition:`` header.
        content_language: ``Content-Language:`` header.
        content_location: ``Content-Location:`` header.
        content_id: ``Content-Id:`` header.
        content_description: ``Content-Description:`` header.
        content_transfer_encoding: ``Content-Transfer-Encoding:`` header.
        body_md5: MD5 hash of the body content.
        size: Size of the message body, in bytes.
        lines: Length of the message body, in lines.

    """

    def __init__(self, subtype: str,
                 content_type_params: Optional[Dict[str, str]],
                 content_disposition: Optional[ContentDispositionHeader],
                 content_language: Optional[UnstructuredHeader],
                 content_location: Optional[UnstructuredHeader],
                 content_id: Optional[UnstructuredHeader],
                 content_description: Optional[UnstructuredHeader],
                 content_transfer_encoding: Optional[UnstructuredHeader],
                 body_md5: Optional[str],
                 size: int, lines: int) -> None:
        super().__init__('text', subtype, content_type_params,
                         content_disposition, content_language,
                         content_location, content_id, content_description,
                         content_transfer_encoding, body_md5, size)
        self.lines = lines

    @property
    def _value(self) -> ListP:
        return ListP([String.build(self.maintype), String.build(self.subtype),
                      _ParamsList(self.content_type_params),
                      String.build(self.content_id),
                      String.build(self.content_description),
                      String.build(self.content_transfer_encoding),
                      Number(self.size), Number(self.lines)])

    @property
    def extended(self) -> ListP:
        """The body structure attributes with extension data."""
        return ListP([String.build(self.maintype), String.build(self.subtype),
                      _ParamsList(self.content_type_params),
                      String.build(self.content_id),
                      String.build(self.content_description),
                      String.build(self.content_transfer_encoding),
                      Number(self.size), Number(self.lines),
                      String.build(self.body_md5),
                      String.build(self.content_disposition),
                      String.build(self.content_language),
                      String.build(self.content_location)])


class MessageBodyStructure(ContentBodyStructure):
    """:class:`BodyStructure` sub-class for ``message/rfc822`` messages.

    Args:
        content_type_params: Parameters from the ``Content-Type:`` header.
        content_disposition: ``Content-Disposition:`` header.
        content_language: ``Content-Language:`` header.
        content_location: ``Content-Location:`` header.
        content_id: ``Content-Id:`` header.
        content_description: ``Content-Description:`` header.
        content_transfer_encoding: ``Content-Transfer-Encoding:`` header.
        body_md5: MD5 hash of the body content.
        size: Size of the message body, in bytes.
        lines: Length of the message body, in lines.
        envelope_structure: Contained message's envelope structure.
        body_structure: Contained message's body structure.

    """

    def __init__(self, content_type_params: Optional[Dict[str, str]],
                 content_disposition: Optional[ContentDispositionHeader],
                 content_language: Optional[UnstructuredHeader],
                 content_location: Optional[UnstructuredHeader],
                 content_id: Optional[UnstructuredHeader],
                 content_description: Optional[UnstructuredHeader],
                 content_transfer_encoding: Optional[UnstructuredHeader],
                 body_md5: Optional[str],
                 size: int, lines: int,
                 envelope_structure: EnvelopeStructure,
                 body_structure: BodyStructure) -> None:
        super().__init__('message', 'rfc822', content_type_params,
                         content_disposition, content_language,
                         content_location, content_id, content_description,
                         content_transfer_encoding, body_md5, size)
        self.lines = lines
        self.envelope_structure = envelope_structure
        self.body_structure = body_structure

    @property
    def _value(self) -> ListP:
        return ListP([String.build(self.maintype), String.build(self.subtype),
                      _ParamsList(self.content_type_params),
                      String.build(self.content_id),
                      String.build(self.content_description),
                      String.build(self.content_transfer_encoding),
                      Number(self.size),
                      self.envelope_structure,
                      self.body_structure,
                      Number(self.lines)])

    @property
    def extended(self) -> ListP:
        """The body structure attributes with extension data."""
        return ListP([String.build(self.maintype), String.build(self.subtype),
                      _ParamsList(self.content_type_params),
                      String.build(self.content_id),
                      String.build(self.content_description),
                      String.build(self.content_transfer_encoding),
                      Number(self.size),
                      self.envelope_structure,
                      self.body_structure.extended,
                      Number(self.lines),
                      String.build(self.body_md5),
                      String.build(self.content_disposition),
                      String.build(self.content_language),
                      String.build(self.content_location)])



"""Utilities for managing a IMAP searches."""

from abc import abstractmethod, ABCMeta
from datetime import datetime
from typing import FrozenSet, Optional, Iterable

from .exceptions import SearchNotAllowed
from .flags import SessionFlags
from .interfaces.message import MessageInterface
from .parsing.specials import SearchKey, SequenceSet
from .parsing.specials.flag import Flag, Answered, Deleted, Draft, Flagged, \
    Recent, Seen
from .selected import SelectedMailbox

__all__ = ['SearchParams', 'SearchCriteria', 'SearchCriteriaSet']


class SearchParams:
    """Defines certain parameters and routines necessary to process any
    kind of search criteria. If a parameter is not supplied, or a method not
    implemented, any search keys that require it will fail.

    Args:
        selected: The active mailbox session.
        max_seq: The highest message sequence ID in the mailbox.
        max_uid: The highest message UID in the mailbox.
        disabled: Search keys that should be disabled.

    """

    __slots__ = ['selected', '_max_seq', '_max_uid', '_disabled']

    def __init__(self, selected: SelectedMailbox,
                 max_seq: int = None, max_uid: int = None,
                 disabled: Iterable[bytes] = None) -> None:
        self.selected = selected
        if max_seq is not None:
            self._max_seq = max_seq
        if max_uid is not None:
            self._max_uid = max_uid
        if disabled is not None:
            self._disabled = frozenset(disabled)

    @property
    def session_flags(self) -> SessionFlags:
        return self.selected.session_flags

    @property
    def max_seq(self) -> int:
        if not hasattr(self, '_max_seq'):
            raise SearchNotAllowed
        return self._max_seq

    @property
    def max_uid(self) -> int:
        if not hasattr(self, '_max_uid'):
            raise SearchNotAllowed
        return self._max_uid

    @property
    def disabled(self) -> FrozenSet[bytes]:
        return getattr(self, '_disabled', frozenset())


class SearchCriteria(metaclass=ABCMeta):
    """Base class for different types of search criteria.

    Args:
        params: The parameters that may be used by some searches.

    """

    def __init__(self, params: SearchParams) -> None:
        self.params = params

    @abstractmethod
    def matches(self, msg_seq: int, msg: MessageInterface) -> bool:
        """Implemented by sub-classes to define the search criteria.

        Args:
            msg_seq: The message sequence ID.
            msg: The message object.

        """
        ...

    @classmethod
    def of(cls, key: SearchKey, params: SearchParams) -> 'SearchCriteria':
        """Factory method for producing a search criteria sub-class from a
        search key.

        Args:
            key: The search key defining the criteria.
            params: The parameters that may be used by some searches.

        """
        if key.value in params.disabled:
            raise SearchNotAllowed(key.value)
        elif key.inverse:
            return InverseSearchCriteria(key.not_inverse, params)
        elif key.value == b'SEQSET':
            return SequenceSetSearchCriteria(key.filter_sequence_set, params)
        elif key.value == b'KEYSET':
            return SearchCriteriaSet(key.filter_key_set, params)
        elif key.value == b'ALL':
            return AllSearchCriteria(params)
        elif key.value == b'OR':
            left_key, right_key = key.filter_key_or
            return OrSearchCriteria(left_key, right_key, params)
        elif key.value == b'ANSWERED':
            return HasFlagSearchCriteria(Answered, True, params)
        elif key.value == b'UNANSWERED':
            return HasFlagSearchCriteria(Answered, False, params)
        elif key.value == b'DELETED':
            return HasFlagSearchCriteria(Deleted, True, params)
        elif key.value == b'UNDELETED':
            return HasFlagSearchCriteria(Deleted, False, params)
        elif key.value == b'DRAFT':
            return HasFlagSearchCriteria(Draft, True, params)
        elif key.value == b'UNDRAFT':
            return HasFlagSearchCriteria(Draft, False, params)
        elif key.value == b'FLAGGED':
            return HasFlagSearchCriteria(Flagged, True, params)
        elif key.value == b'UNFLAGGED':
            return HasFlagSearchCriteria(Flagged, False, params)
        elif key.value == b'RECENT':
            return HasFlagSearchCriteria(Recent, True, params)
        elif key.value == b'OLD':
            return HasFlagSearchCriteria(Recent, False, params)
        elif key.value == b'SEEN':
            return HasFlagSearchCriteria(Seen, True, params)
        elif key.value == b'UNSEEN':
            return HasFlagSearchCriteria(Seen, False, params)
        elif key.value == b'KEYWORD':
            return HasFlagSearchCriteria(key.filter_flag, True, params)
        elif key.value == b'UNKEYWORD':
            return HasFlagSearchCriteria(key.filter_flag, False, params)
        elif key.value == b'NEW':
            return NewSearchCriteria(params)
        elif key.value == b'BEFORE':
            return DateSearchCriteria(key.filter_datetime, '<', params)
        elif key.value == b'ON':
            return DateSearchCriteria(key.filter_datetime, '=', params)
        elif key.value == b'SINCE':
            return DateSearchCriteria(key.filter_datetime, '>', params)
        elif key.value == b'SENTBEFORE':
            return HeaderDateSearchCriteria(key.filter_datetime, '<', params)
        elif key.value == b'SENTON':
            return HeaderDateSearchCriteria(key.filter_datetime, '=', params)
        elif key.value == b'SENTSINCE':
            return HeaderDateSearchCriteria(key.filter_datetime, '>', params)
        elif key.value == b'SMALLER':
            return SizeSearchCriteria(key.filter_int, '<', params)
        elif key.value == b'LARGER':
            return SizeSearchCriteria(key.filter_int, '>', params)
        elif key.value in (b'BCC', b'CC', b'FROM', b'SUBJECT', b'TO'):
            return EnvelopeSearchCriteria(key.value, key.filter_str, params)
        elif key.value == b'HEADER':
            name, value = key.filter_header
            return HeaderSearchCriteria(name, value, params)
        elif key.value in (b'BODY', b'TEXT'):
            value = key.filter_str
            return BodySearchCriteria(value, key.value == b'TEXT', params)
        raise SearchNotAllowed(key.value)


class SearchCriteriaSet(SearchCriteria):
    """Search criteria composed of a set of search criteria that must all
    match. If the set is empty, nothing will match.

    Args:
        keys: The set of search keys that must match.
        params: The parameters that may be used by some searches.

    """

    def __init__(self, keys: FrozenSet[SearchKey],
                 params: SearchParams) -> None:
        super().__init__(params)
        self.all_criteria = [SearchCriteria.of(key, params) for key in keys]

    def matches(self, msg_seq: int, msg: MessageInterface) -> bool:
        """The message matches if all the defined search key criteria match.

        Args:
            msg_seq: The message sequence ID.
            msg: The message object.

        """
        return all(crit.matches(msg_seq, msg) for crit in self.all_criteria)


class InverseSearchCriteria(SearchCriteria):
    """Matches only if the given search criteria does not match."""

    def __init__(self, key: SearchKey, params: SearchParams) -> None:
        super().__init__(params)
        self.key = SearchCriteria.of(key, params)

    def matches(self, msg_seq: int, msg: MessageInterface) -> bool:
        return not self.key.matches(msg_seq, msg)


class AllSearchCriteria(SearchCriteria):
    """Always matches anything."""

    def matches(self, msg_seq: int, msg: MessageInterface):
        return True


class OrSearchCriteria(SearchCriteria):
    """Matches if either of the search criteria match."""

    def __init__(self, left: SearchKey, right: SearchKey,
                 params: SearchParams) -> None:
        super().__init__(params)
        self.left = SearchCriteria.of(left, self.params)
        self.right = SearchCriteria.of(right, self.params)

    def matches(self, msg_seq: int, msg: MessageInterface) -> bool:
        return (self.left.matches(msg_seq, msg)
                or self.right.matches(msg_seq, msg))


class SequenceSetSearchCriteria(SearchCriteria):
    """Matches if the message is contained in the sequence set."""

    def __init__(self, seq_set: SequenceSet, params: SearchParams) -> None:
        super().__init__(params)
        self.seq_set = seq_set

    def matches(self, msg_seq: int, msg: MessageInterface) -> bool:
        if self.seq_set.uid:
            return self.seq_set.contains(msg.uid, self.params.max_uid)
        else:
            return self.seq_set.contains(msg_seq, self.params.max_seq)


class HasFlagSearchCriteria(SearchCriteria):
    """Matches if the message has the given flag in their permanent or
    session flag sets.

    """

    def __init__(self, flag: Flag, expected: bool,
                 params: SearchParams) -> None:
        super().__init__(params)
        self.flag = flag
        self.expected = expected

    def matches(self, msg_seq: int, msg: MessageInterface) -> bool:
        has_flag = self.flag in msg.get_flags(self.params.session_flags)
        expected = self.expected
        return (has_flag and expected) or (not expected and not has_flag)


class NewSearchCriteria(SearchCriteria):
    """Matches if the message is considered "new", i.e. recent and unseen."""

    def matches(self, msg_seq: int, msg: MessageInterface) -> bool:
        flags = msg.get_flags(self.params.session_flags)
        return Recent in flags and Seen not in flags


class DateSearchCriteria(SearchCriteria):
    """Matches by comparing against the internal date of the message."""

    def __init__(self, when: datetime, cmp: str, params: SearchParams) -> None:
        super().__init__(params)
        self.when = when.date()
        self.cmp = cmp

    @classmethod
    def _get_msg_date(cls, msg: MessageInterface) -> Optional[datetime]:
        return msg.internal_date

    def matches(self, msg_seq: int, msg: MessageInterface) -> bool:
        msg_datetime = self._get_msg_date(msg)
        if msg_datetime is None:
            return False
        msg_date = msg_datetime.date()
        if self.cmp == '<':  # BEFORE
            return msg_date < self.when
        elif self.cmp == '=':  # ON
            return msg_date == self.when
        elif self.cmp == '>':  # SINCE
            return msg_date > self.when
        raise ValueError(self.cmp)


class HeaderDateSearchCriteria(DateSearchCriteria):
    """Matches by comparing against the ``Date:`` header of the message."""

    @classmethod
    def _get_msg_date(cls, msg: MessageInterface) -> Optional[datetime]:
        envelope = msg.get_envelope_structure()
        return envelope.date.datetime if envelope.date else None


class SizeSearchCriteria(SearchCriteria):
    """Matches by comparing against the size of the message."""

    def __init__(self, size: int, cmp: str, params: SearchParams) -> None:
        super().__init__(params)
        self.size = size
        self.cmp = cmp

    def matches(self, msg_seq: int, msg: MessageInterface) -> bool:
        size = msg.get_size()
        if self.cmp == '<':
            return size < self.size
        elif self.cmp == '>':
            return size > self.size
        raise ValueError(self.cmp)


class EnvelopeSearchCriteria(SearchCriteria):
    """Matches by checking for strings within various fields of the envelope
    structure.

    """

    def __init__(self, key: bytes, value: str, params: SearchParams) -> None:
        super().__init__(params)
        self.key = key
        self.value = value

    def matches(self, msg_seq: int, msg: MessageInterface) -> bool:
        envelope = msg.get_envelope_structure()
        if self.key == b'BCC':
            if not envelope.bcc:
                return False
            return any(self.value in str(bcc) for bcc in envelope.bcc)
        elif self.key == b'CC':
            if not envelope.cc:
                return False
            return any(self.value in str(cc) for cc in envelope.cc)
        elif self.key == b'FROM':
            if not envelope.from_:
                return False
            return any(self.value in str(from_) for from_ in envelope.from_)
        elif self.key == b'SUBJECT':
            if not envelope.subject:
                return False
            return self.value in str(envelope.subject)
        elif self.key == b'TO':
            if not envelope.to:
                return False
            return any(self.value in str(to) for to in envelope.to)
        raise ValueError(self.key)


class HeaderSearchCriteria(SearchCriteria):
    """Matches if the message has a header containing a value."""

    def __init__(self, name: str, value: str, params: SearchParams) -> None:
        super().__init__(params)
        self.name = name
        self.value = value

    def matches(self, msg_seq: int, msg: MessageInterface) -> bool:
        values = msg.get_header(self.name)
        return any(self.value in str(value) for value in values)


class BodySearchCriteria(SearchCriteria):
    """Matches if the message body contains a value."""

    def __init__(self, value: str, with_headers: bool,
                 params: SearchParams) -> None:
        super().__init__(params)
        self.with_headers = with_headers
        self.value = bytes(value, 'utf-8', 'replace')

    def matches(self, msg_seq: int, msg: MessageInterface) -> bool:
        data = msg.get_body() if self.with_headers else msg.get_text()
        return data is not None and self.value in data



"""Defines socket interaction, accepting connections and handling a basic
command/response flow.

"""

import asyncio
import binascii
import re
from asyncio import IncompleteReadError, StreamReader, StreamWriter
from base64 import b64encode, b64decode
from ssl import SSLContext
from typing import List, Optional

from pysasl import ServerChallenge, AuthenticationError, \
    AuthenticationCredentials

from .bytes import as_memoryview
from .concurrent import Event, TimeoutError
from .config import IMAPConfig
from .context import current_command, socket_info
from .exceptions import ResponseError
from .interfaces.session import LoginProtocol
from .parsing.command import Command
from .parsing.commands import Commands
from .parsing.command.nonauth import AuthenticateCommand, LoginCommand, \
    StartTLSCommand
from .parsing.command.select import IdleCommand
from .parsing.exceptions import RequiresContinuation, BadCommand
from .parsing.response import ResponseContinuation, Response, ResponseCode, \
    ResponseBad, ResponseNo, ResponseBye, ResponseOk
from .sockinfo import SocketInfo
from .state import ConnectionState

__all__ = ['IMAPServer', 'IMAPConnection']


class Disconnected(Exception):
    """Thrown if the remote socket closes when the server expected input."""
    pass


class IMAPServer:
    """Callable object that creates and runs :class:`IMAPConnection` objects
    when :func:`asyncio.start_server` receives a new connection.

    Args:
        login: Login callback that takes authentication credentials and
            returns a :class:`~pymap.interfaces.session.SessionInterface`
            object.
        config: Settings to use for the IMAP server.

    """

    __slots__ = ['commands', 'login', 'config']

    def __init__(self, login: LoginProtocol, config: IMAPConfig) -> None:
        super().__init__()
        self.commands = config.commands
        self.login = login
        self.config = config

    async def __call__(self, reader: StreamReader,
                       writer: StreamWriter) -> None:
        conn = IMAPConnection(self.commands, self.config, reader, writer)
        state = ConnectionState(self.login, self.config)
        await conn.run(state)


class IMAPConnection:
    """Runs a single IMAP connection from start to finish.

    Args:
        commands: Defines the IMAP commands available to the connection.
        config: Settings to use for the IMAP connection.
        reader: The input stream for the socket.
        writer: The output stream for the socket.

    """

    __slots__ = ['commands', 'config', 'params', 'bad_command_limit',
                 '_print', 'reader', 'writer']

    def __init__(self, commands: Commands, config: IMAPConfig,
                 reader: StreamReader,
                 writer: StreamWriter) -> None:
        super().__init__()
        self.commands = commands
        self.config = config
        self.params = config.parsing_params
        self.bad_command_limit = config.bad_command_limit
        self._reset_streams(reader, writer)
        self._print = self._real_print if config.debug else self._noop_print

    def _reset_streams(self, reader: StreamReader, writer: StreamWriter):
        self.reader = reader
        self.writer = writer
        socket_info.set(SocketInfo(writer))

    @classmethod
    def _real_print(cls, prefix: str, output: bytes) -> None:
        prefix = prefix % socket_info.get().socket.fileno()
        lines = re.split(br'\r?\n', output)
        if not lines[-1]:
            lines = lines[:-1]
        for line in lines:
            line_str = str(line, 'utf-8', 'replace')
            print(prefix, line_str)

    @classmethod
    def _noop_print(cls, prefix: str, output: bytes) -> None:
        pass

    async def read_continuation(self, literal_length: int) -> bytes:
        try:
            extra_literal = await self.reader.readexactly(literal_length)
        except IncompleteReadError:
            raise Disconnected
        extra_line: bytes = await self.reader.readline()
        if self.reader.at_eof():
            raise Disconnected
        extra = extra_literal + extra_line
        self._print('%d -->|', extra)
        return as_memoryview(extra)

    async def authenticate(self, state: ConnectionState, mech_name: bytes) \
            -> Optional[AuthenticationCredentials]:
        mech = state.auth.get_server(mech_name)
        if not mech:
            return None
        responses: List[ServerChallenge] = []
        while True:
            try:
                return mech.server_attempt(responses)
            except ServerChallenge as chal:
                chal_bytes = b64encode(chal.get_challenge())
                cont = ResponseContinuation(chal_bytes)
                await self.write_response(cont)
                resp_bytes = await self.read_continuation(0)
                try:
                    chal.set_response(b64decode(resp_bytes))
                except binascii.Error as exc:
                    raise AuthenticationError(exc)
                if bytes(resp_bytes).rstrip(b'\r\n') == b'*':
                    raise AuthenticationError('Authentication canceled.')
                responses.append(chal)

    async def read_command(self) -> Command:
        line = as_memoryview(await self.reader.readline())
        if self.reader.at_eof():
            raise Disconnected
        self._print('%d -->|', line)
        conts: List[bytes] = []
        while True:
            try:
                cmd, _ = self.commands.parse(
                    line, self.params.copy(continuations=conts.copy()))
            except RequiresContinuation as req:
                cont = ResponseContinuation(req.message)
                await self.write_response(cont)
                ret = await self.read_continuation(req.literal_length)
                conts.append(ret)
            else:
                return cmd

    async def read_idle_done(self, cmd: IdleCommand) -> None:
        buf = await self.read_continuation(len(cmd.continuation))
        cmd.parse_done(buf, self.params.copy(tag=cmd.tag))

    async def write_response(self, resp: Response) -> None:
        raw = bytes(resp)
        self.writer.write(raw)
        await self.writer.drain()
        self._print('%d <--|', raw)

    async def start_tls(self, ssl_context: SSLContext) -> None:
        loop = asyncio.get_event_loop()
        transport = self.writer.transport
        protocol = transport.get_protocol()  # type: ignore
        new_transport = await loop.start_tls(  # type: ignore
            transport, protocol, ssl_context, server_side=True)
        protocol._stream_reader = StreamReader(loop=loop)
        protocol._client_connected_cb = self._reset_streams
        protocol.connection_made(new_transport)
        self._print('%d <->|', b'<TLS handshake>')

    async def send_error_disconnect(self) -> None:
        resp = ResponseBye(b'Unhandled server error.')
        try:
            await self.write_response(resp)
            self.writer.close()
        except IOError:
            pass

    async def write_bad_command(self, bad: BadCommand,
                                bad_commands: int) -> bool:
        resp = bad.get_response()
        if self.bad_command_limit and bad_commands >= self.bad_command_limit:
            bye = ResponseBye(b'Too many errors, disconnecting.')
            resp.add_untagged(bye)
        await self.write_response(resp)
        return resp.is_terminal

    async def write_updates(self, state: ConnectionState,
                            cmd: IdleCommand, done: Event) -> None:
        async for resp in state.receive_updates(cmd, done):
            await self.write_response(resp)

    async def idle(self, state: ConnectionState, cmd: IdleCommand) -> Response:
        response = await state.do_command(cmd)
        if not isinstance(response, ResponseOk):
            return response
        done = self.config.new_event()
        await self.write_response(ResponseContinuation(b'Idling.'))
        task = asyncio.create_task(self.write_updates(state, cmd, done))
        try:
            await self.read_idle_done(cmd)
        except BadCommand as bad:
            return bad.get_response()
        finally:
            done.set()
        await task
        return response

    async def run(self, state: ConnectionState) -> None:
        """Start the socket communication with the IMAP greeting, and then
        enter the command/response cycle.

        Args:
            state: Defines the interaction with the backend plugin.

        """
        self._print('%d +++|', bytes(socket_info.get()))
        bad_commands = 0
        try:
            greeting = await state.do_greeting()
        except ResponseError as exc:
            resp = exc.get_response(b'*')
            resp.condition = ResponseBye.condition
            await self.write_response(resp)
            self.writer.close()
            return
        else:
            await self.write_response(greeting)
        while True:
            try:
                cmd = await self.read_command()
            except BadCommand as bad:
                bad_commands += 1
                if await self.write_bad_command(bad, bad_commands):
                    break
            except (ConnectionResetError, BrokenPipeError):
                break
            except Disconnected:
                break
            except Exception:
                await self.send_error_disconnect()
                raise
            else:
                bad_commands = 0
                prev_cmd = current_command.set(cmd)
                try:
                    if isinstance(cmd, AuthenticateCommand):
                        creds = await self.authenticate(state, cmd.mech_name)
                        response = await state.do_authenticate(cmd, creds)
                    elif isinstance(cmd, LoginCommand):
                        creds = AuthenticationCredentials(
                            cmd.userid.decode('utf-8', 'surrogateescape'),
                            cmd.password.decode('utf-8', 'surrogateescape'))
                        response = await state.do_login(cmd, creds)
                    elif isinstance(cmd, IdleCommand):
                        response = await self.idle(state, cmd)
                    else:
                        response = await state.do_command(cmd)
                except ResponseError as exc:
                    resp = exc.get_response(cmd.tag)
                    await self.write_response(resp)
                    if resp.is_terminal:
                        break
                except AuthenticationError as exc:
                    msg = bytes(str(exc), 'utf-8', 'surrogateescape')
                    resp = ResponseBad(cmd.tag, msg)
                    await self.write_response(resp)
                except TimeoutError:
                    resp = ResponseNo(cmd.tag, b'Operation timed out.',
                                      ResponseCode.of(b'TIMEOUT'))
                    await self.write_response(resp)
                except Exception:
                    await self.send_error_disconnect()
                    raise
                else:
                    await self.write_response(response)
                    if response.is_terminal:
                        break
                    if isinstance(cmd, StartTLSCommand) and state.ssl_context \
                            and isinstance(response, ResponseOk):
                        await self.start_tls(state.ssl_context)
                finally:
                    current_command.reset(prev_cmd)
        self._print('%d ---|', b'<disconnected>')
        self.writer.close()




import re
from typing import Tuple, Optional, List, Mapping, Iterable

from . import AString, SequenceSet
from .. import Params, Parseable
from ..exceptions import NotParseable
from ..primitives import Number, ListP
from ...bytes import BytesFormat

__all__ = ['ExtensionOption', 'ExtensionOptions']


class ExtensionOption(Parseable[bytes]):
    """Represents a single command option, which may or may not have an
    associated value.

    See Also:
        `RFC 4466 2.1. <https://tools.ietf.org/html/rfc4466#section-2.1>`_

    Args:
        option: The name of the option.
        arg: The option argument, if any.

    """

    _opt_pattern = re.compile(br'[a-zA-Z_.-][a-zA-Z0-9_.:-]*')

    def __init__(self, option: bytes, arg: ListP) -> None:
        super().__init__()
        self.option = option
        self.arg = arg
        self._raw_arg: Optional[bytes] = None

    @property
    def value(self) -> bytes:
        return self.option

    def __bytes__(self) -> bytes:
        if self.arg.value:
            return BytesFormat(b'%b %b') % (self.option, self.raw_arg)
        else:
            return self.option

    @property
    def raw_arg(self) -> bytes:
        if self._raw_arg is None:
            if not self.arg:
                self._raw_arg = b''
            elif len(self.arg) == 1:
                arg_0 = self.arg.value[0]
                if isinstance(arg_0, (Number, SequenceSet)):
                    self._raw_arg = bytes(arg_0)
                else:
                    self._raw_arg = bytes(self.arg)
            else:
                self._raw_arg = bytes(self.arg)
        return self._raw_arg

    @classmethod
    def _parse_arg(cls, buf: bytes, params: Params) \
            -> Tuple[ListP, bytes]:
        try:
            num, buf = Number.parse(buf, params)
        except NotParseable:
            pass
        else:
            arg = ListP([num])
            return arg, buf
        try:
            seq_set, buf = SequenceSet.parse(buf, params)
        except NotParseable:
            pass
        else:
            arg = ListP([seq_set])
            return arg, buf
        try:
            params_copy = params.copy(list_expected=[AString, ListP])
            return ListP.parse(buf, params_copy)
        except NotParseable:
            pass
        return ListP([]), buf

    @classmethod
    def parse(cls, buf: bytes, params: 'Params') \
            -> Tuple['ExtensionOption', bytes]:
        start = cls._whitespace_length(buf)
        match = cls._opt_pattern.match(buf, start)
        if not match:
            raise NotParseable(buf[start:])
        option = match.group(0).upper()
        buf = buf[match.end(0):]
        arg, buf = cls._parse_arg(buf, params)
        return cls(option, arg), buf


class ExtensionOptions(Parseable[Mapping[bytes, ListP]]):
    """Represents a set of command options, which may or may not have an
    associated argument. Command options are always optional, so the parsing
    will not fail, it will just return an empty object.

    See Also:
        `RFC 4466 2.1. <https://tools.ietf.org/html/rfc4466#section-2.1>`_

    Args:
        options: The mapping of options to argument.

    """

    _opt_pattern = re.compile(br'[a-zA-Z_.-][a-zA-Z0-9_.:-]*')
    _empty: Optional['ExtensionOptions'] = None

    def __init__(self, options: Iterable[ExtensionOption]) -> None:
        super().__init__()
        self.options: Mapping[bytes, ListP] = \
            {opt.option: opt.arg for opt in options}
        self._raw: Optional[bytes] = None

    @classmethod
    def empty(cls) -> 'ExtensionOptions':
        if cls._empty is None:
            cls._empty = ExtensionOptions({})
        return cls._empty

    @property
    def value(self) -> Mapping[bytes, ListP]:
        return self.options

    def has(self, option: bytes) -> bool:
        return option in self.options

    def get(self, option: bytes) -> Optional[ListP]:
        return self.options.get(option, None)

    def __len__(self) -> int:
        return len(self.options)

    def __bytes__(self) -> bytes:
        if self._raw is None:
            parts = [ExtensionOption(option, arg)
                     for option, arg in sorted(self.options.items())]
            self._raw = b'(' + BytesFormat(b' ').join(parts) + b')'
        return self._raw

    @classmethod
    def _parse_paren(cls, buf: bytes, paren: bytes) -> bytes:
        start = cls._whitespace_length(buf)
        if buf[start:start + 1] != paren:
            raise NotParseable(buf)
        return buf[start + 1:]

    @classmethod
    def _parse(cls, buf: bytes, params: Params) \
            -> Tuple['ExtensionOptions', bytes]:
        buf = cls._parse_paren(buf, b'(')
        result: List[ExtensionOption] = []
        while True:
            try:
                option, buf = ExtensionOption.parse(buf, params)
            except NotParseable:
                break
            else:
                result.append(option)
        buf = cls._parse_paren(buf, b')')
        return cls(result), buf

    @classmethod
    def parse(cls, buf: bytes, params: 'Params') \
            -> Tuple['ExtensionOptions', bytes]:
        try:
            return cls._parse(buf, params)
        except NotParseable:
            return cls.empty(), buf




from collections import OrderedDict
from io import BytesIO
from typing import Tuple, Sequence, Dict, Optional, Iterable, AsyncIterable

from pymap.concurrent import ReadWriteLock
from pymap.exceptions import MailboxNotFound, MailboxConflict
from pymap.mailbox import MailboxSnapshot
from pymap.message import AppendMessage
from pymap.parsing.specials import FetchRequirement
from pymap.selected import SelectedSet, SelectedMailbox

from ..mailbox import Message, MailboxDataInterface, MailboxSetInterface

__all__ = ['Message', 'MailboxData', 'MailboxSet']


class MailboxData(MailboxDataInterface[Message]):
    """Implementation of :class:`~pymap.backend.mailbox.MailboxDataInterface`
    for the dict backend.

    """

    def __init__(self, name: str) -> None:
        self._name = name
        self._readonly = False
        self._messages_lock = ReadWriteLock.for_asyncio()
        self._selected_set = SelectedSet()
        self._reset_messages()

    def _reset_messages(self) -> None:
        self._uid_validity = MailboxSnapshot.new_uid_validity()
        self._max_uid = 100
        self._messages: Dict[int, Message] = OrderedDict()

    @property
    def name(self) -> str:
        return self._name

    @property
    def readonly(self) -> bool:
        return self._readonly

    @property
    def uid_validity(self) -> int:
        return self._uid_validity

    @property
    def next_uid(self) -> int:
        return self._max_uid + 1

    @property
    def messages_lock(self) -> ReadWriteLock:
        return self._messages_lock

    @property
    def selected_set(self) -> SelectedSet:
        return self._selected_set

    def parse_message(self, append_msg: AppendMessage) -> Message:
        msg_data = BytesIO(append_msg.message)
        return Message.parse(0, msg_data, append_msg.flag_set,
                             append_msg.when, recent=True)

    async def add(self, message: Message, recent: bool = False) -> Message:
        async with self.messages_lock.write_lock():
            self._max_uid += 1
            msg_copy = message.copy(self._max_uid)
            msg_copy.recent = recent
            self._messages[msg_copy.uid] = msg_copy
            return msg_copy

    async def get(self, uid: int,
                  requirement: FetchRequirement = FetchRequirement.METADATA) \
            -> Optional[Message]:
        async with self.messages_lock.read_lock():
            ret = self._messages.get(uid)
            if ret is None and uid <= self._max_uid:
                ret = Message(uid, expunged=True)
            return ret

    async def delete(self, uids: Iterable[int]) -> None:
        async with self.messages_lock.write_lock():
            for uid in uids:
                try:
                    del self._messages[uid]
                except KeyError:
                    pass

    async def claim_recent(self, selected: SelectedMailbox) -> None:
        async for msg in self.messages():
            if msg.recent:
                msg.recent = False
                selected.session_flags.add_recent(msg.uid)

    async def save_flags(self, messages: Iterable[Message]) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    async def uids(self) -> AsyncIterable[int]:
        async with self.messages_lock.read_lock():
            for key in self._messages.keys():
                yield key

    async def messages(self) -> AsyncIterable[Message]:
        async with self.messages_lock.read_lock():
            for msg in self._messages.values():
                yield msg

    async def items(self) -> AsyncIterable[Tuple[int, Message]]:
        async with self.messages_lock.read_lock():
            for key, msg in self._messages.items():
                yield (key, msg)


class MailboxSet(MailboxSetInterface[MailboxData]):
    """Implementation of :class:`~pymap.backend.mailbox.MailboxSetInterface`
    for the dict backend.

    """

    def __init__(self) -> None:
        super().__init__()
        self._inbox = MailboxData('INBOX')
        self._set: Dict[str, 'MailboxData'] = OrderedDict()
        self._set_lock = ReadWriteLock.for_asyncio()
        self._subscribed: Dict[str, bool] = {}

    @property
    def inbox(self) -> MailboxData:
        return self._inbox

    @property
    def delimiter(self) -> str:
        return '.'

    async def set_subscribed(self, name: str, subscribed: bool) -> None:
        async with self._set_lock.write_lock():
            self._subscribed[name] = subscribed

    async def list_subscribed(self) -> Sequence[str]:
        async with self._set_lock.read_lock():
            return [child for child in self._set.keys()
                    if self._subscribed.get(child)]

    async def list_mailboxes(self) -> Sequence[str]:
        async with self._set_lock.read_lock():
            return list(self._set.keys())

    async def get_mailbox(self, name: str,
                          try_create: bool = False) -> 'MailboxData':
        if name.upper() == 'INBOX':
            return self.inbox
        async with self._set_lock.read_lock():
            if name not in self._set:
                raise MailboxNotFound(name, try_create)
            return self._set[name]

    async def add_mailbox(self, name: str) -> 'MailboxData':
        async with self._set_lock.read_lock():
            if name in self._set:
                raise MailboxConflict(name)
        async with self._set_lock.write_lock():
            self._set[name] = ret = MailboxData(name)
        return ret

    async def delete_mailbox(self, name: str) -> None:
        async with self._set_lock.read_lock():
            if name not in self._set:
                raise MailboxNotFound(name)
        async with self._set_lock.write_lock():
            del self._set[name]

    async def rename_mailbox(self, before: str, after: str) -> 'MailboxData':
        async with self._set_lock.read_lock():
            if before != 'INBOX' and before not in self._set:
                raise MailboxNotFound(before)
            elif after in self._set:
                raise MailboxConflict(after)
        if before == 'INBOX':
            async with self._set_lock.write_lock():
                self._set[after] = ret = MailboxData(after)
                ret._uid_validity = self.inbox._uid_validity
                ret._max_uid = self.inbox._max_uid
                ret._messages = self.inbox._messages
                self.inbox._reset_messages()
                return ret
        else:
            async with self._set_lock.write_lock():
                self._set[after] = self._set[before]
                del self._set[before]
                return self._set[after]



"""IMAP server with pluggable Python backends."""

import asyncio
import traceback
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

from pkg_resources import iter_entry_points

from .core import __version__
from .server import IMAPServer


def _load_backends(parser):
    ret = {}
    for entry_point in iter_entry_points('pymap.backend'):
        try:
            mod = entry_point.load()
        except ImportError:
            traceback.print_exc()
            parser.exit(1, 'Error importing registered backend: %s\n'
                        % entry_point.name)
        else:
            ret[entry_point.name] = mod
    return ret


def main():
    parser = ArgumentParser(description=__doc__,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('--debug', action='store_true',
                        help='increase printed output for debugging')
    parser.add_argument('--version', action='version',
                        version='%(prog)s' + __version__)
    subparsers = parser.add_subparsers(dest='backend',
                                       help='which pymap backend to use')
    listener = parser.add_argument_group('server arguments')
    listener.add_argument('--port', action='store', type=int, default=1143,
                          help='the port to listen on')
    listener.add_argument('--cert', action='store', help='cert file for TLS')
    listener.add_argument('--key', action='store', help='key file for TLS')
    listener.add_argument('--insecure-login', action='store_true',
                          help='allow plaintext login without TLS')

    backends = _load_backends(parser)

    for mod in backends.values():
        mod.add_subparser(subparsers)
    args = parser.parse_args()

    try:
        backend = backends[args.backend]
    except KeyError:
        parser.error('Expected backend name.')
        return

    loop = asyncio.get_event_loop()
    login_func, config = loop.run_until_complete(backend.init(args))
    callback = IMAPServer(login_func, config)
    coro = asyncio.start_server(callback, port=args.port, loop=loop)
    server = loop.run_until_complete(coro)

    try:
        loop.run_forever()
    except KeyboardInterrupt:
        print()

    server.close()
    loop.run_until_complete(server.wait_closed())
    loop.close()
    return 0


if __name__ == '__main__':
    main()
